################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import yaml
import logging
from pathlib import Path
from collections import Counter
from typing import Optional, Dict, Any

import torch
import lightning as L
from hydra.utils import instantiate
from torchmetrics import MetricCollection


from .select_model import select_model
from .lora import from_lora_to_merged_model, to_lora_model
from ..dataset.quantity import get_target_property, get_embedding_property
from ..utils.metrics import build_metrics, update_metrics
from .. utils._global import DTYPE, DEVICE


# === LightningModule wrap ===
class LightningWrapperModel(L.LightningModule):
    def __init__(
            self, 
            cfg: Dict, 
            model: torch.nn.Module,
            target_property: list[str], 
            embedding_property: list[str], 
            statistics,  # TODO
        ):
        super().__init__()
        self.save_hyperparameters(ignore=["model"])
        self.cfg = cfg
        self.statistics = statistics  # TODO
        model = from_lora_to_merged_model(model)
        self.model = to_lora_model(cfg.get('finetune', {}), model)

        # === Loss === 
        self.loss_fn = instantiate(cfg["loss"])  # TODO
        self.loss_property = target_property

        # === Metric ===
        self._create_metrics("train")
        self._create_metrics("val")
        synth_metric = cfg.get('synth_metric', None)
        if synth_metric:
            total = sum([v for k, v in synth_metric.items() if k != 'monitor_metric_name'])
            if total:
                synth_metric = {
                    k: (v / total if k != 'monitor_metric_name' else v)
                    for k, v in synth_metric.items()
                }
        self.synth_metric = synth_metric
        test_sets = cfg.get("dataset", {}).get("test_files", [])
        if test_sets is None:
            self.num_test_sets = 0
        elif isinstance(test_sets, str):
            self.num_test_sets = 1
            self._create_metrics(f"test_0")
        else:
            self.num_test_sets = len(test_sets)
            for i in range(self.num_test_sets):
                self._create_metrics(f"test_{i}")

        # === Misc ===
        self.force_dtype = DTYPE[cfg.get("dataset", {}).get("force_dtype", None)]
        self.no_valid_set = cfg.get("dataset", {}).get("no_valid_set", False)

    def setup(self, stage: Optional[str] = None):
        self.sync_dist = self.trainer.num_devices > 1
        if self.ema is not None:
            self.ema.to(self.device)

    def _create_metrics(self, prefix):
        metric_collection = MetricCollection(build_metrics(prefix, self.loss_property))
        setattr(self, f"{prefix}_metrics", metric_collection)

    def _process_batch(self, batch):
        output = self.model(batch)
        loss = self.loss_fn(output, batch)
        return output, loss

    def _shared_step(self, batch, batch_idx, prefix):
        if self.force_dtype is not None:
            batch = batch.apply(
                lambda x: x.to(self.force_dtype) if x.is_floating_point() else x
            )            
        output, loss = self._process_batch(batch)

        self.log(
            f"{prefix}/loss",
            loss,
            prog_bar=True,
            logger=True,
            on_epoch=True,
            batch_size=len(batch),
            on_step=True,
            sync_dist=self.sync_dist,
            reduce_fx="mean",
        )

        metrics = getattr(self, f"{prefix}_metrics")
        update_metrics(metrics, prefix, output, batch, self.loss_property)

        return loss

    def training_step(self, batch, batch_idx):
        return self._shared_step(batch, batch_idx, "train")

    def validation_step(self, batch, batch_idx):
        with torch.enable_grad():
            output = self._shared_step(batch, batch_idx, "val")
        return output

    def test_step(self, batch, batch_idx, dataloader_idx=0):
        with torch.enable_grad():
            output = self._shared_step(batch, batch_idx, f"test_{dataloader_idx}")
        return output

    def on_epoch_end(self, prefix):
        metrics = getattr(self, f"{prefix}_metrics").compute()
        for name, metric in metrics.items():
            self.log(
                name,
                metric,
                sync_dist=self.sync_dist,
                reduce_fx="mean",
                on_epoch=True,
                add_dataloader_idx=True,
            )
        logging.info("═" * 50 + "\n")

        if prefix == 'val':
            if self.synth_metric is not None:
                synth_metric = sum([metrics[k] * v * 0.001 for k, v in self.synth_metric.items() if k != 'monitor_metric_name'])
                self.log(
                    f"{prefix}/synth_metric",
                    synth_metric,
                    sync_dist=self.sync_dist,
                    reduce_fx="mean",
                    on_epoch=True,
                    add_dataloader_idx=True,
                )
        getattr(self, f"{prefix}_metrics").reset()
    def on_train_epoch_end(self):
        self.on_epoch_end("train")

    def on_train_epoch_start(self):
        lr = self.trainer.optimizers[0].param_groups[0]["lr"]
        logging.info(f"LR: {lr:.6e}")

    def on_validation_epoch_end(self):
        self.on_epoch_end("val")

    def on_test_epoch_end(self):
        logging.info(f"The error is 1000 times")
        for i in range(self.num_test_sets):
            metrics = getattr(self, f"test_{i}_metrics").compute()
            for name, metric in metrics.items():
                self.log(
                    name,
                    metric,
                    sync_dist=self.sync_dist,
                    reduce_fx="mean",
                    on_epoch=True,
                    add_dataloader_idx=True,
                )
                logging.info(f"{name}: {metric.item():.6f}")
            getattr(self, f"test_{i}_metrics").reset()
            logging.info("═" * 50 + "\n")

    def on_fit_start(self):
        if hasattr(self, "ema") and self.ema is not None:
            self.ema.to(self.device)

    def setup(self, stage=None):
        if self.trainer.world_size > 1:
            self.sync_dist = True
        else:
            self.sync_dist = False

    def configure_optimizers(self):
        optimizer = instantiate(
            {**{k: v for k, v in self.cfg["optimizer"].items() if k != "extra"}},
            params=self.parameters(),
        )
        if "scheduler" in self.cfg:
            lr_scheduler_cfg = self.cfg.get("scheduler", {}).get("extra", {})
            monitor = lr_scheduler_cfg.get("monitor", "val/loss")
            interval = lr_scheduler_cfg.get("interval", "epoch")
            frequency = lr_scheduler_cfg.get("frequency", 1)
            scheduler = instantiate(
                {**{k: v for k, v in self.cfg["scheduler"].items() if k != "extra"}},
                optimizer=optimizer,
            )
            if self.no_valid_set:
                return {
                    "optimizer": optimizer,
                    "lr_scheduler": {
                        "scheduler": scheduler,
                        "interval": interval,
                        "frequency": frequency,
                    },
                }
            else:
                return {
                    "optimizer": optimizer,
                    "lr_scheduler": {
                        "scheduler": scheduler,
                        "monitor": monitor,
                        "interval": interval,
                        "frequency": frequency,
                    },
                }
        else:
            return {
                "optimizer": optimizer,
            }

    @classmethod
    def load_from_checkpoint(
        cls,
        ckpt_path: str,
        map_location: str = "cpu",
        strict: Optional[bool] = True,
        use_ema: int | bool = 1,
    ) -> Any:

        checkpoint = torch.load(
            ckpt_path, map_location=map_location, weights_only=False
        )
        dtypes = [
            v.dtype for v in checkpoint["state_dict"].values()
            if hasattr(v, "dtype") and torch.is_floating_point(v)
        ]
        dominant_dtype = Counter(dtypes).most_common(1)[0][0] # original training precision
        cfg = checkpoint['hyper_parameters']['cfg']
        target_property = get_target_property(cfg)
        embedding_property = get_embedding_property(cfg)
        statistics = checkpoint['hyper_parameters']['statistics']
        model = select_model(cfg, statistics, target_property, embedding_property)
        model = to_lora_model(cfg.get('finetune', {}), model)
        model.to(dtype=dominant_dtype)
        state_dict = {
            k[len("model.") :]: v for k, v in checkpoint["state_dict"].items() if k.startswith("model.")
        }
        model.load_state_dict(state_dict, strict=strict)

        # === SWA ===
        if 'swa' in cfg['callbacks']:
            logging.debug("Since swa is enabled, skip trying load ema.")
            return model.to(map_location)
        
        # === EMA ===
        if bool(use_ema) and "ema_state_dict" in checkpoint:
            ema_params = checkpoint['ema_state_dict']['shadow_params']
            for idx, (name, _ ) in enumerate(model.named_parameters()):
                state_dict[name] = ema_params[idx]
            model.load_state_dict(state_dict, strict=strict)

        return model.to(map_location)

def _load_tace(
    model: str | Path | torch.nn.Module,
    device: Optional[str | torch.device] = None,
    strict: Optional[bool] = True,
    use_ema: bool = True,
    **kwargs: Any,
):
    device = DEVICE[device]
    if isinstance(model, str | Path):
        model_path = str(model)
        if model_path.endswith(".ckpt"):
            model = LightningWrapperModel.load_from_checkpoint(
                model_path,
                map_location=device,
                strict=strict,
                use_ema=use_ema,
            )
        elif (model_path.endswith(".pt") or model_path.endswith(".pth")):
            obj = torch.load(
                model_path,
                map_location=device,
                weights_only=False,
                **kwargs,
            )
            if isinstance(obj, dict) and "state_dict" in obj:
                model = select_model(
                    **{k: v for k, v in obj.items() if k != "state_dict"}
                )
                model.load_state_dict(obj["state_dict"])
            elif isinstance(obj, torch.nn.Module):
                model = obj
        else:
            raise ValueError("❌ Model path must end with .ckpt, .pt or .pth")
    elif isinstance(model, torch.nn.Module):
        pass
    else:
        raise TypeError("Model must be a path or torch.nn.Module")

    return model.to(device)

def load_tace(
    model: str | Path | torch.nn.Module,
    device: Optional[str | torch.device] = None,
    strict: Optional[bool] = True,
    use_ema: bool = True,
    target_property: Optional[list[str]] = None,
    **kwargs: Any,
):
    model = _load_tace(
        model=model,
        device=device,
        strict=strict,
        use_ema=use_ema,
        **kwargs,
    )

    if target_property:
        model.reset_target_property(target_property)

    return model

def finetune(cfg: Dict) -> torch.nn.Module:

    model = load_tace(
        cfg["finetune_from_model"],
        device='cpu',
        strict=True,
        use_ema=True,
    )

    precision = cfg['trainer']['precision']
    dtype = DTYPE.get(precision, torch.float64)

    if dtype == torch.float16:
        model.half()      
    elif dtype == torch.float32:
        model.float()   
    else:
        model.double()

    logging.info(f"Load model for Fine-tunning")
    
    return model




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import shutil
import logging
from pathlib import Path
from urllib.parse import urlparse
from collections.abc import Mapping


from huggingface_hub import hf_hub_download


from ..utils._global import CACHE_DIR


OAM_SERIES = {
    "TACE-v1-OMat24-M": "https://huggingface.co/xvzemin/tace-foundations/resolve/main/TACE-v1-OMat24-M.pt",
    "TACE-v1-OAM-M": "https://huggingface.co/xvzemin/tace-foundations/resolve/main/TACE-v1-OAM-M.pt",
}

REICO_SERIES = {
    # "TACE-v1-REICO-PdAgCHO": "https://huggingface.co/xvzemin/tace-foundations/resolve/main/TACE-v1-REICO-5-PdAgCHO.pt",
}


def _parse_hf_resolve_url(url: str) -> tuple[str, str, str]:
    """
    Parse Hugging Face resolve URL into (repo_id, revision, filename)

    Example:
    https://huggingface.co/xvzemin/tace-oam/resolve/main/foo.pt
    -> ("xvzemin/tace-oam", "main", "foo.pt")
    """
    parsed = urlparse(url)
    parts = parsed.path.strip("/").split("/")

    try:
        resolve_idx = parts.index("resolve")
    except ValueError:
        raise ValueError(f"Not a valid Hugging Face resolve URL: {url}")

    repo_id = "/".join(parts[:resolve_idx])
    revision = parts[resolve_idx + 1]
    filename = "/".join(parts[resolve_idx + 2:])

    if not repo_id or not revision or not filename:
        raise ValueError(f"Malformed Hugging Face resolve URL: {url}")

    return repo_id, revision, filename


class CachedModelRegistry(Mapping):
    def __init__(self, registry: dict[str, str]):
        self._registry = registry

    def __getitem__(self, key: str) -> Path:
        if key not in self._registry:
            print(f"[ERROR], Unknown pretrained model: {key}")
            self.print_models()
            raise KeyError(key)

        url = self._registry[key]
        target = Path(CACHE_DIR) / Path(urlparse(url).path).name
        
        if target.exists():
            return Path(target)
        
        # download
        try:
            repo_id, revision, filename = _parse_hf_resolve_url(url)
            path = hf_hub_download(
                repo_id=repo_id,
                filename=filename,
                revision=revision,
                cache_dir=CACHE_DIR,
            )
        except Exception as e:
            raise RuntimeError(
                f"Failed to download pretrained model.\n"
                f"URL:\n"
                f"  {url}\n"
                f"Please manually download and place it at:\n"
                f"  {target}"
            ) from e
        
        # copy
        shutil.copy2(path, target)

        return Path(target)

    def __iter__(self):
        return iter(self._registry)

    def __len__(self):
        return len(self._registry)

    def list_models(self) -> list[str]:
        return sorted(self._registry.keys())

    def print_models(self):
        print("Available pretrained models:")
        for name in self.list_models():
            print(f"  - {name}")


tace_foundations = CachedModelRegistry(
    OAM_SERIES | REICO_SERIES
)



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import List, Dict


import torch
from torch import nn, Tensor

from .linear import Linear
from .gate import NormGate, GatedGate


def format_list(obj, ndigits=4):
    if isinstance(obj, int):
        return str(obj)
    elif isinstance(obj, float):
        return f"{obj:.{ndigits}f}"
    elif isinstance(obj, (list, tuple)):
        return "[" + ", ".join(format_list(x, ndigits) for x in obj) + "]"
    else:
        return str(obj)
  
  
class OneHotToAtomicEnergy(torch.nn.Module):
    def __init__(self, atomic_energies: List[Dict[int, float]]) -> None:
        super().__init__()
        assert atomic_energies is not None
        atomic_energy_list = []
        for atomic_energy in atomic_energies:
            atomic_energy_list.append(
                [float(v) for _, v in atomic_energy.items()]
            )
        self.register_buffer(
            "atomic_energy",
            torch.tensor(
                atomic_energy_list,
                dtype=torch.get_default_dtype(),
            ),
        )

    def forward(self, x: Tensor) -> Tensor: 
        return torch.matmul(x, self.atomic_energy.T)

    def __repr__(self):
        s = f"{self.__class__.__name__}(\n"
        s += "  atomic_energies = {\n"

        data = self.atomic_energy.detach().cpu().numpy()

        for i in range(data.shape[0]):
            s += f"    level {i}: {format_list(data[i].tolist(), 4)}\n"

        s += "  }\n"
        s += ")"
        return s


class ScaleShift(torch.nn.Module):
    def __init__(
        self,
        scale_dicts: List[Dict[int, float]] = [],
        shift_dicts: List[Dict[int, float]] = [],
        scale_trainable: bool = False,
        shift_trainable: bool = False,
    ):
        super().__init__()

        self.has_scale = len(scale_dicts) > 0
        self.has_shift = len(shift_dicts) > 0
        self.num_levels = max(len(scale_dicts), len(shift_dicts))
        atomic_numbers = sorted(
            set().union(*[d.keys() for d in scale_dicts] if scale_dicts else [])
            | set().union(*[d.keys() for d in shift_dicts] if shift_dicts else [])
        )
        self.register_buffer(
            "atomic_numbers", torch.tensor(atomic_numbers, dtype=torch.int64)
        )

        if self.has_scale:
            scale_list = []
            for d in scale_dicts:
                scale_list.append([d.get(z, 1.0) for z in atomic_numbers])
            scale_tensor = torch.tensor(scale_list, dtype=torch.get_default_dtype())
            if scale_trainable:
                self.scale = nn.Parameter(scale_tensor)
            else:
                self.register_buffer("scale", scale_tensor)

        if self.has_shift:
            shift_list = []
            for d in shift_dicts:
                shift_list.append([d.get(z, 0.0) for z in atomic_numbers])
            shift_tensor = torch.tensor(shift_list, dtype=torch.get_default_dtype())
            if shift_trainable:
                self.shift = nn.Parameter(shift_tensor)
            else:
                self.register_buffer("shift", shift_tensor)

    def forward(self, node_energy, node_attrs, ptr, edge_index, batch, node_level):
        if not (self.has_scale or self.has_shift):
            return node_energy

        num_graphs = ptr.numel() - 1
        num_nodes = ptr[1:] - ptr[:-1]

        if edge_index.numel() == 0:
            num_edges = torch.zeros(num_graphs, dtype=torch.int64, device=node_energy.device)
        else:
            edge_batch = batch[edge_index[1]]
            num_edges = torch.bincount(edge_batch, minlength=num_graphs)

        isolated_mask = (num_nodes == 1) & (num_edges == 0)

        if self.has_scale:
            node_scale = (node_attrs * self.scale[node_level]).sum(dim=-1)
            if isolated_mask.any():
                isolated_nodes = torch.isin(batch, torch.where(isolated_mask)[0])
                node_scale[isolated_nodes] = 0.0
            node_energy = node_energy * node_scale

        if self.has_shift:
            node_shift = (node_attrs * self.shift[node_level]).sum(dim=-1)
            if isolated_mask.any():
                isolated_nodes = torch.isin(batch, torch.where(isolated_mask)[0])
                node_shift[isolated_nodes] = 0.0
            node_energy = node_energy + node_shift

        return node_energy

    def __repr__(self):

        s = f"{self.__class__.__name__}(\n"
        s += f"  atomic_numbers = {self.atomic_numbers.tolist()}\n"

        if self.has_scale:
            s += "  scale = {\n"
            for lvl in range(self.scale.shape[0]):
                data = self.scale[lvl].detach().cpu().numpy().tolist()
                s += f"    level {lvl}: {format_list(data, 4)}\n"
            s += "  }\n"
        else:
            s += "  scale = None\n"

        if self.has_shift:
            s += "  shift = {\n"
            for lvl in range(self.shift.shape[0]):
                data = self.shift[lvl].detach().cpu().numpy().tolist()
                s += f"    level {lvl}: {format_list(data, 4)}\n"
            s += "  }\n"
        else:
            s += "  shift = None\n"

        s += ")"
        return s
    
    @classmethod
    def build_from_config(cls, statistics, cfg: Dict):
        required_keys = [
            "scale_type",
            "shift_type",
            "scale_trainable",
            "shift_trainable",
            "scale_dict",
            "shift_dict",
        ]
        assert all(
            k in cfg for k in required_keys
        ), f"Missing keys in scale_shift config: {required_keys}"

        scale_key = cfg["scale_type"]
        shift_key = cfg["shift_type"]

        scale_dicts = []
        shift_dicts = []

        for stats in statistics:
            scale_stat = {z: 1.0 for z in stats["atomic_numbers"]}
            shift_stat = {z: 0.0 for z in stats["atomic_numbers"]}

            if scale_key is not None:
                assert hasattr(stats, scale_key), f"{scale_key} not found in statistics"
                scale_stat = getattr(stats, scale_key, scale_stat)

            if shift_key is not None:
                assert hasattr(stats, shift_key), f"{shift_key} not found in statistics"
                shift_stat = getattr(stats, shift_key, shift_stat)

            scale_dict = {int(k): float(v) for k, v in scale_stat.items()}
            shift_dict = {int(k): float(v) for k, v in shift_stat.items()}

            if isinstance(cfg["scale_dict"], list):
                if len(cfg["scale_dict"]) > len(scale_dicts):
                    for k, v in cfg["scale_dict"][len(scale_dicts)].items():
                        k = int(k)
                        assert k in scale_dict, f"Invalid atomic number {k} in scale_dict"
                        scale_dict[k] = float(v)

            if isinstance(cfg["shift_dict"], list):
                if len(cfg["shift_dict"]) > len(shift_dicts):
                    for k, v in cfg["shift_dict"][len(shift_dicts)].items():
                        k = int(k)
                        assert k in shift_dict, f"Invalid atomic number {k} in shift_dict"
                        shift_dict[k] = float(v)

            scale_dicts.append(scale_dict)
            shift_dicts.append(shift_dict)

        return cls(
            scale_dicts=scale_dicts,
            shift_dicts=shift_dicts,
            scale_trainable=cfg["scale_trainable"],
            shift_trainable=cfg["shift_trainable"],
        )


class NormNonlinearity(torch.nn.Module):
    def __init__(
        self,
        rmax: int,
        in_dim: int,
        gate: str = "silu",
    ) -> None:
        super().__init__()
        self.rmax = rmax
        self.in_dim = in_dim
        self.gate = gate
        self.gates = torch.nn.ModuleList(
            [NormGate[gate](r, in_dim) for r in range(rmax + 1)]
        )

    def forward(self, in_dict: Dict[int, torch.Tensor]) -> Dict[int, torch.Tensor]:
        out_dict = {}
        for r, gate in enumerate(self.gates):
            if r in in_dict:
                out_dict[r] = gate(in_dict[r])
        return out_dict

    def __repr__(self):
        return f"{self.__class__.__name__}(rmax={self.rmax}, channel={self.in_dim}, gate={self.gate})"
    
    
class GatedNonlinearity(torch.nn.Module):
    def __init__(
        self,
        rmax: int,
        in_dim: int,
        gate: str = "silu",
    ) -> None:
        super().__init__()
        self.rmax = rmax
        self.in_dim = in_dim
        self.gate = gate
        self.gates = torch.nn.ModuleList(
            [GatedGate(r, gate) for r in range(rmax + 1)]
        )

        self.linears_1 = torch.nn.ModuleList()
        for r in range(rmax+1):
            out_dim = in_dim * (rmax+1) if r == 0 else in_dim
            self.linears_1.append(
                Linear(
                    in_dim=in_dim,
                    out_dim=out_dim,
                    bias=False,
                    l=r,
                )
            )

    def forward(self, in_dict: Dict[int, torch.Tensor]) -> Dict[int, torch.Tensor]:

        out_dict = {}
        gates = self.linears_1[0](in_dict[0])
        gate = gates[:, 0:self.in_dim]
        out_dict[0] = self.gates[0](gate, gate)

        start = self.in_dim
        for r in range(1, self.rmax+1):
            stop = start+self.in_dim
            gate = gates[:, start:stop]
            out_dict[r] = self.gates[r](gate, self.linears_1[r](in_dict[r]))
            start=stop

        return out_dict

    def __repr__(self):
        return f"{self.__class__.__name__}(rmax={self.rmax}, channel={self.in_dim}, gate={self.gate})"


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import warnings
from typing import Optional


import torch
from ase import units
from ase.calculators.calculator import Calculator, all_changes
from ase.calculators.mixing import SumCalculator
from torch_geometric.loader import DataLoader


from ...lightning import load_tace
from ...dataset.quantity import PROPERTY
from ...dataset.element import TorchElement
from ...dataset.graph import from_atoms
from ...dataset.quantity import (
    PROPERTY,
    KEYS,
    KeySpecification,
    update_keyspec_from_kwargs,
)
from ...utils._global import DTYPE, DEVICE


class TACEAseCalc(Calculator):
    """
    Initialize a TACEAseCalc. We support the most fundamental potential energy surface property and multi-fidelity, 
    multi-head, etc. For some advanced features, you need to store the attributes that need to be embedded in atoms.info, 
    atoms.arrays or add a funciton by yourself. If you only need to predict, you can directly use the `tace-eval` 
    command. It will output the predicted files, and if you add the `--test` option, it will also output the errors.

    Parameters
    ----------
    model_path : str
        Path to the trained model, file ends with pt, .pth or .ckpt.
    device : str | torch.device, optional
        The device to run computations on, e.g., cpu or cuda.
        If None, the device is automatically inferred.
    dtype : str, optional, default=None
        Model dtype for computations, e.g., float32 or float64.
    level : int
        Specify which fidelity level to use. 
    spin_on : bool
        If your model uses spin_on uie embedding, you can control whether 
        your calculation enables spin polarization.
    target_property: list(str)
        Extra caculate hessians, atomic_virials, Conservative polarizability, etc,
        If you want to use this parameter, you must provide all the required physical quantities.
    neighborlist_backend: str
        Support backend in one of [ase, matscipy, vesin], recommend matscipy
    **kwargs
        Additional keyword arguments passed to the ASE Calculator base class.
    """

    def __init__(
        self,
        model: str,
        *,
        dtype: Optional[str] = None,
        device: Optional[str] = None,
        level: Optional[int] = None,
        spin_on: Optional[bool] = None,
        target_property: Optional[list[str]] = None,
        neighborlist_backend: str = "matscipy",
        **kwargs,
    ):
        super().__init__(**kwargs)
        # === init ===
        model = load_tace(
            model, 
            device, 
            strict=True, 
            use_ema=True, 
            target_property=target_property
        )
        model_dtype = model.readout_fn.cutoff.dtype
        dtype = dtype or model_dtype
        self.dtype = DTYPE[dtype]
        self.device = DEVICE[device or torch.device("cuda" if torch.cuda.is_available() else "cpu")]
        torch.set_default_dtype(self.dtype)
        if DTYPE[dtype] != DTYPE[model_dtype]:
            warnings.warn(
                f"Model dtype {model_dtype} != default dtype {dtype}. "
                f"This may cause silent type conversions."
            )
        model = model.to(dtype=self.dtype)
        self.target_property = list(set(model.target_property))
        self.embedding_property = model.embedding_property
        self.implemented_properties = []
        for p in self.target_property:
            ase_name = PROPERTY[p]['ase_name']
            save_name = ase_name if ase_name else p
            if save_name == 'energy':
                self.implemented_properties.extend(["energy" ,"free_energy"])
            else:
                self.implemented_properties.append(save_name)
        self.max_neighbors = getattr(model.readout_fn, "max_neighbors", None)
        self.cutoff = float(model.readout_fn.cutoff.item())
        self.element = TorchElement([int(z) for z in model.readout_fn.atomic_numbers.cpu().tolist()])
        self.neighborlist_backend = neighborlist_backend
        model.eval()
        for param in model.parameters():
            param.requires_grad = False

        self.keySpecification = KeySpecification()
        update_keyspec_from_kwargs(self.keySpecification, KEYS)
 
        if level is not None:
            self.level = level
            model.reset_computing_level(level) 
        else:
            self.level = model.get_computing_level()

        if spin_on is not None:
            self.spin_on = spin_on
            model.reset_spin_on(spin_on) 
        else:
            self.spin_on = model.get_spin_on() 

        self.model = model.to(self.device)

    def calculate(self, atoms=None, properties=None, system_changes=all_changes):
        Calculator.calculate(self, atoms)
        atoms.info['level'] = self.level # fidelity level
        # === dataloader ===
        data = [
            from_atoms(
                self.element,
                atoms,
                self.cutoff,
                max_neighbors=self.max_neighbors,
                target_property=self.target_property,
                embedding_property=self.embedding_property,
                keyspec=self.keySpecification,
                universal_embedding=self.model.universal_embedding,
                training=False,
                neighborlist_backend=self.neighborlist_backend,
            ) 
        ]
        dataloader = DataLoader(
            dataset=data,
            batch_size=1,
            shuffle=False,
            drop_last=False,
        )

        batch = next(iter(dataloader)).to(self.device)
   
        # === forward ===
        outs = self.model(batch)
        # === update ===
        self.results = {}
        for p in self.target_property:
            p_rank = PROPERTY[p]['rank']
            p_scope = PROPERTY[p]['scope']
            ase_name = PROPERTY[p]['ase_name']
            save_name = ase_name if ase_name else p
            if p_scope == 'per-system':
                if p_rank == 0:
                    if p == 'energy':
                        energy = outs[p].detach().cpu().item()
                        self.results['energy'] = energy
                        self.results["free_energy"] = self.results['energy']
                    else:
                        self.results[save_name] = outs[p].detach().cpu().item()
                else:
                    self.results[save_name] = outs[p].detach().cpu().numpy().squeeze(0)
            elif p_scope == 'per-atom':
                self.results[save_name] = outs[p].detach().cpu().numpy()
            elif p_scope == 'per-edge':
                self.results[save_name] = outs[p].detach().cpu().numpy()
            else:
                self.results[save_name] = outs[p].detach().cpu().numpy()

def add_dispersion(
    base_calc: Calculator,
    damping: str = "bj",  # choices: ["zero", "bj", "zerom", "bjm"]
    dispersion_xc: str = "pbe",
    dispersion_cutoff: float = 40.0 * units.Bohr,
    **kwargs,
) -> SumCalculator:
    try:
        from torch_dftd.torch_dftd3_calculator import TorchDFTD3Calculator
    except ImportError as e:
        raise RuntimeError(
            "Please install torch-dftd to use dispersion corrections (see https://github.com/pfnet-research/torch-dftd)"
        ) from e
    
    d3_calc = TorchDFTD3Calculator(
        dtype=base_calc.dtype,
        device=base_calc.device,
        damping=damping,
        xc=dispersion_xc,
        cutoff=dispersion_cutoff,
        **kwargs,
    )
    return SumCalculator([base_calc, d3_calc])


# TODO not for users now
################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import math
import heapq
from typing import List, Iterator, Optional

import torch
import torch.distributed as dist
from torch.utils.data import BatchSampler, Dataset
from torch_geometric.data import Data


class EdgeBalancedBatchSampler(BatchSampler):
    """
    Batch sampler for Graph Datasets that balances batches by total number of edges.

    Features:
    - Builds global batches by max_edges_per_batch.
    - Splits each batch across GPUs using Largest Processing Time (LPT) greedy algorithm.
    - Supports single-GPU (non-DDP) and multi-GPU (DDP).
    """

    def __init__(
        self,
        dataset: Dataset,
        max_edges_per_batch: int,
        shuffle: bool = True,
        drop_last: bool = False,
        seed: int = 42,
    ):
        self.dataset = dataset
        self.max_edges_per_batch = int(max_edges_per_batch)
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.seed = seed

        # --- handle single GPU vs DDP ---
        if dist.is_available() and dist.is_initialized():
            self.rank = dist.get_rank()
            self.world_size = dist.get_world_size()
        else:
            self.rank = 0
            self.world_size = 1  # single GPU fallback

        self.epoch = 0

        # Lazy cache for edge counts
        self._edge_counts_cache: List[Optional[int]] = [None] * len(dataset)

    def set_epoch(self, epoch: int):
        """Set current training epoch (affects shuffle)."""
        self.epoch = epoch

    def __iter__(self) -> Iterator[List[int]]:
        indices = list(range(len(self.dataset)))
        if self.shuffle:
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(indices), generator=g).tolist()

        # Build global batches by max_edges_per_batch
        global_batches = self._build_global_batches(indices)

        for batch_indices in global_batches:
            # Partition each global batch across GPUs using LPT
            per_rank_batches = self._partition_lpt(batch_indices)

            local_batch = per_rank_batches[self.rank]

            if self.drop_last and len(local_batch) == 0:
                continue

            yield local_batch

    def __len__(self) -> int:
        total_edges = sum(self._edge_count(i) for i in range(len(self.dataset)))
        return math.ceil(total_edges / self.max_edges_per_batch)

    def _edge_count(self, idx: int) -> int:
        """Lazy compute edge count with caching."""
        if self._edge_counts_cache[idx] is None:
            graph = self.dataset[idx]
            if hasattr(graph, "edge_index"):
                w = int(graph.edge_index.size(1))
            elif hasattr(graph, "num_edges"):
                w = int(graph.num_edges)
            else:
                raise AttributeError(f"Sample {idx} has no `edge_index` or `num_edges` attribute")
            self._edge_counts_cache[idx] = w
        return self._edge_counts_cache[idx]

    def _build_global_batches(self, indices: List[int]) -> List[List[int]]:
        """Accumulate samples into global batches by max_edges_per_batch."""
        batches = []
        current = []
        current_edges = 0

        for idx in indices:
            w = self._edge_count(idx)

            # handle huge graph
            if w > self.max_edges_per_batch:
                if not self.drop_last:
                    batches.append([idx])
                continue

            if current_edges + w > self.max_edges_per_batch:
                if current or not self.drop_last:
                    batches.append(current)
                current = []
                current_edges = 0

            current.append(idx)
            current_edges += w

        if current and not self.drop_last:
            batches.append(current)

        return batches

    def _partition_lpt(self, batch_indices: List[int]) -> List[List[int]]:
        """Partition a global batch into per-rank sub-batches using LPT greedy algorithm."""
        batch_indices_sorted = sorted(batch_indices, key=lambda i: self._edge_count(i), reverse=True)

        # Heap: (load, rank_index)
        heap = [(0, i) for i in range(self.world_size)]
        heapq.heapify(heap)

        buckets: List[List[int]] = [[] for _ in range(self.world_size)]
        for idx in batch_indices_sorted:
            load, rank_i = heapq.heappop(heap)
            buckets[rank_i].append(idx)
            load += self._edge_count(idx)
            heapq.heappush(heap, (load, rank_i))

        return buckets


class ListDataset(Dataset):
    def __init__(self, data_list):
        self.data_list = data_list

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx):
        return self.data_list[idx]



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import argparse
import logging
import sys

from ..foundations import tace_foundations


def download_one(registry, name: str):
    path = registry[name]
    logging.info(f"Model '{name}' is ready at: {path}")


def download_all(registry):
    logging.info("Downloading all available models...")
    for name in registry.list_models():
        path = registry[name]
        logging.info(f"✔ {name} -> {path}")


def main():
    parser = argparse.ArgumentParser(
        description="Download pretrained TACE models"
    )

    parser.add_argument(
        "-m",
        "--model",
        type=str,
        help="Name of the model to download (download all if omitted)",
    )

    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List all available models and exit",
    )

    args = parser.parse_args()

    logging.basicConfig(
        level=logging.INFO,
        format="[%(levelname)s] %(message)s",
    )

    registry = tace_foundations

    if args.list:
        print("Available pretrained models:")
        for name in registry.list_models():
            print(f"  - {name}")
        sys.exit(0)

    if args.model is None:
        download_all(registry)
    else:
        download_one(registry, args.model)


if __name__ == "__main__":
    main()



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
'''
This file is about the radial relatied operations.

The code framework and most of the implementation are modified from MACE.
https://github.com/ACEsuit/mace

In addition, we provide several extra operations related to the radial components.

MIT License

Copyright (c) 2022 ACEsuit/mace

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in the 
Software without restriction, including without limitation the rights to use, copy, 
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
'''
import math
from typing import List


import ase.data
import numpy as np
import torch
from torch import Tensor
from omegaconf import ListConfig
from scipy.optimize import brentq
from scipy.special import jv


from ...utils.torch_scatter import scatter_sum


def jn_taylor(n, x, terms=5):
    x = x.to(torch.float64)
    result = torch.zeros_like(x)
    for k in range(terms):
        coeff = ((-1) ** k) / (math.factorial(k) * math.factorial(2 * k + n + 1))
        term = coeff * x ** (2 * k + n)
        result += term
    result /= 2**n
    return result


def compute_jn_zeros(n: int, k: int) -> np.ndarray:
    def spherical_bessel_jn(r, order):
        return np.sqrt(np.pi / (2 * r)) * jv(
            order + 0.5, r
        )  # from first bessel to first spherical bessel

    zeros = []
    guess_points = np.arange(1, k + 20) * np.pi

    found = 0
    i = 0
    while found < k and i < len(guess_points) - 1:
        a, b = guess_points[i], guess_points[i + 1]
        try:
            root = brentq(
                spherical_bessel_jn, a, b, args=(n)
            )  # search roots for spherical_bessel_jn in [a, b]
            zeros.append(root)
            found += 1
        except ValueError:
            pass
        i += 1
    return np.array(zeros, dtype=np.float64)


class j0_SphericalBesselBasis(torch.nn.Module):
    """The Bessel Basis is proposed in the DimeNet: https://www.cs.cit.tum.de/daml/dimenet/"""

    def __init__(
        self, cutoff: float = 6.0, num_basis: int = 8, trainable: bool = False
    ) -> None:
        super().__init__()
        self.num_basis = num_basis
        bessel_roots = (
            math.pi
            / cutoff
            * torch.linspace(
                start=1.0,
                end=num_basis,
                steps=num_basis,
                dtype=torch.get_default_dtype(),
            )
        )
        if trainable:
            self.bessel_weights = torch.nn.Parameter(bessel_roots)
        else:
            self.register_buffer("bessel_weights", bessel_roots)

        self.register_buffer(
            "cutoff", torch.tensor(cutoff, dtype=torch.get_default_dtype())
        )
        self.register_buffer(
            "prefactor",
            torch.tensor(math.sqrt(2.0 / cutoff), dtype=torch.get_default_dtype()),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # [..., 1]
        numerator = torch.sin(self.bessel_weights * x)  # [..., num_basis]
        return self.prefactor * (numerator / x)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(cutoff={self.cutoff}, num_basis={len(self.bessel_weights)}, "
            f"trainable={self.bessel_weights.requires_grad})"
        )


class jn_SphericalBesselBasis(torch.nn.Module):
    "arbitrary order n >= 0"

    def __init__(
        self,
        cutoff: float = 6.0,
        order: int | List[int] = 0,
        num_basis: int | List[int] = 8,
        trainable: bool = False,
    ) -> None:
        super().__init__()

        # assert torch.get_default_dtype() == torch.float64, f"This class requires float64 precision, but got {torch.get_default_dtype()}"

        num_zero = num_basis
        if isinstance(order, int):
            if order < 0:
                raise ValueError("order must be a nonnegative integer")
            order = [order]
        else:
            if not isinstance(order, (List, ListConfig)):
                raise TypeError("order must be a list of nonnegative integer")
            if not all(isinstance(x, int) and x >= 0 for x in order):
                raise ValueError("All elements of order must be nonnegative integer")

        if isinstance(num_zero, int):
            if num_zero <= 0:
                raise ValueError("num_zero must be a positive integer")
            num_zero = [num_zero]
        else:
            if not isinstance(order, (List, ListConfig)):
                raise TypeError("num_zero must be a list of positive integers")
            if not all(isinstance(x, int) and x > 0 for x in num_zero):
                raise ValueError("All elements of num_zero must be positive integers")

        if len(order) != len(num_zero):
            raise ValueError(
                f"order and num_zero must have the same length, "
                f"but got {len(order)} and {len(num_zero)}"
            )

        zeros = []
        for o, n in zip(order, num_zero):
            zeros.append(compute_jn_zeros(o, n))

        normalizer = self._compute_normalizer(cutoff, order, zeros)

        self.register_buffer(
            "normalizer",
            torch.tensor(
                [y for x in normalizer for y in x], dtype=torch.get_default_dtype()
            ).unsqueeze(0),
        )  # (1, sum(order*zeros))
        if trainable:
            self.zeros = torch.nn.Parameter(
                torch.tensor(
                    [y for x in zeros for y in x], dtype=torch.get_default_dtype()
                ).unsqueeze(0)
            )
        else:
            self.register_buffer(
                "zeros",
                torch.tensor(
                    [y for x in zeros for y in x], dtype=torch.get_default_dtype()
                ).unsqueeze(0),
            )
        self.register_buffer(
            "cutoff", torch.tensor(cutoff, dtype=torch.get_default_dtype())
        )
        self.order = order
        self.num_zero = num_zero

    def torch_jn(self, order, x):
        if order == 0:
            return torch.sin(x) / x
        elif order == 1:
            return torch.sin(x) / x**2 - torch.cos(x) / x
        elif order == 2:
            return (3 / x**3 - 1 / x) * torch.sin(x) - (3 * torch.cos(x) / x**2)
        elif order == 3:
            return (15 / x**4 - 6 / x**2) * torch.sin(x) - (
                15 / x**3 - 1 / x
            ) * torch.cos(x)
        elif order == 4:
            return (105 / x**5 - 45 / x**3 + 1 / x) * torch.sin(x) - (
                105 / x**4 - 10 / x**2
            ) * torch.cos(x)
        elif order == 5:
            return (945 / x**6 - 420 / x**4 + 15 / x**2) * torch.sin(x) - (
                945 / x**5 - 105 / x**3 + 1 / x
            ) * torch.cos(x)
        elif order == 6:
            return (10395 / x**7 - 4725 / x**5 + 210 / x**3 - 1 / x) * torch.sin(
                x
            ) - (10395 / x**6 - 1260 / x**4 + 21 / x**2) * torch.cos(x)
        elif order == 7:
            return (
                135135 / x**8 - 62370 / x**6 + 3150 / x**4 - 28 / x**2
            ) * torch.sin(x) - (
                135135 / x**7 - 17325 / x**5 + 378 / x**3 - 1 / x
            ) * torch.cos(
                x
            )
        elif order == 8:
            return (
                2027025 / x**9 - 945945 / x**7 + 51975 / x**5 - 630 / x**3 + 1 / x
            ) * torch.sin(x) - (
                2027025 / x**8 - 270270 / x**6 + 6930 / x**4 - 36 / x**2
            ) * torch.cos(
                x
            )
        elif order == 9:
            return (
                34459425 / x**10
                - 16216200 / x**8
                + 945945 / x**6
                - 13860 / x**4
                + 45 / x**2
            ) * torch.sin(x) - (
                34459425 / x**9 - 4729725 / x**7 + 135135 / x**5 - 990 / x**3 + 1 / x
            ) * torch.cos(
                x
            )
        else:

            N = (
                order + 100
            )  # Starting point for backward recursion (Miller's algorithm)
            device, dtype = x.device, x.dtype
            j = torch.zeros(N + 1, *x.shape, dtype=dtype, device=device)

            j[N] = 1e-40
            j[N - 1] = 0.0

            for n in range(N - 1, 0, -1):
                j[n - 1] = (2 * n + 1) / x * j[n] - j[n + 1]

            # Normalize using accurately computed j0
            j0_ground_truth = torch.sin(x) / x
            scale_factor = j0_ground_truth / j[0]  # shape: (...,)
            j_normalized = j * scale_factor.unsqueeze(0)

            return j_normalized[order]

    def forward(self, r: Tensor):  # [..., 1]
        orig_dtype = r.dtype
        r = r.to(torch.float64)
        cutoff = self.cutoff.to(torch.float64)
        zeros = self.zeros.to(torch.float64)
        normalizer = self.normalizer.to(torch.float64)

        r = zeros * (r / cutoff)

        basis = []
        idx = 0
        for o, n in zip(self.order, self.num_zero):
            r_order = r[..., idx : idx + n]
            jn = self.torch_jn(o, r_order)
            basis.append(jn)
            idx += n
        basis = torch.cat(basis, dim=-1)

        out = basis * normalizer

        return out.to(orig_dtype)

    def _compute_normalizer(self, cutoff, order, zeros):
        normalizer = []
        for o, zero in zip(order, zeros):
            o_normalizer = []
            for i in range(len(zero)):
                z = zero[i]
                j_next = np.sqrt(np.pi / (2 * z)) * jv(o + 1 + 0.5, z)
                norm = np.sqrt(2 / (cutoff**3 * j_next**2))
                o_normalizer.append(norm)
            normalizer.append(o_normalizer)
        return normalizer

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(cutoff={self.cutoff}, order={self.order},  num_zero={self.num_zero}, "
            f"trainable={self.zeros.requires_grad})"  # zeros={self.zeros.tolist()}
        )


class Normalized_j0_SphericalBesselBasis(torch.nn.Module):
    """
    From Allergo https://www.nature.com/articles/s41467-023-36329-y
    """

    def __init__(
        self,
        r_max: float,
        r_min: float = 0.0,
        original_basis=j0_SphericalBesselBasis,
        original_basis_kwargs: dict = {},
        n: int = 4000,  # number of integral grid points
        norm_basis_mean_shift: bool = True,  # calculate the mean value and offset it; otherwise, only standardize it
    ):
        super().__init__()
        self.basis = original_basis(**original_basis_kwargs)
        self.r_min = r_min
        self.r_max = r_max
        assert self.r_min >= 0.0
        assert self.r_max > r_min
        self.n = n
        self.num_basis = self.basis.num_basis

        # Uniform distribution on [r_min, r_max)
        with torch.no_grad():
            # don't take 0 in case of weirdness like bessel at 0 # weirdness 不可思议的 离奇的 singular point?
            rs = torch.linspace(r_min, r_max, n + 1)[1:]
            rs.unsqueeze_(-1)
            bs = self.basis(rs)  # TODO for xzm
            assert bs.ndim == 2 and len(bs) == n
            if norm_basis_mean_shift:
                basis_std, basis_mean = torch.std_mean(bs, dim=0)
            else:
                basis_std = bs.square().mean().sqrt()  # rms
                basis_mean = torch.as_tensor(
                    0.0, device=basis_std.device, dtype=basis_std.dtype
                )

        self.register_buffer("_mean", basis_mean)
        self.register_buffer("_inv_std", torch.reciprocal(basis_std))

    def forward(self, x: torch.Tensor) -> Tensor:
        return (self.basis(x) - self._mean) * self._inv_std


class CosineCutoff(torch.nn.Module):
    def __init__(self, cutoff: float) -> None:
        super().__init__()
        self.register_buffer(
            "cutoff",
            torch.tensor(cutoff, dtype=torch.get_default_dtype()),
        )

    def forward(self, length: torch.Tensor) -> torch.Tensor:
        return torch.square(
            0.5 * torch.cos(length * (torch.pi / self.cutoff)) + 0.5
        ).unsqueeze(-1)
    

class PolynomialCutoff(torch.nn.Module):
    """Envelope function(PolynomialCutoff funciton) is proposed in the DimeNet: https://www.cs.cit.tum.de/daml/dimenet/"""

    def __init__(self, cutoff: float, p: float = 6.0, start: float = 0.0) -> None:
        super().__init__()
        assert p > 0.1, f"In PolynomialCutoff function p ({p}) should greater than 0.1"
        assert (
            start < cutoff
        ), f"In cutoff function start ({start}) should less than cutoff ({cutoff})"
        self.register_buffer("p", torch.tensor(p, dtype=torch.get_default_dtype()))
        self.register_buffer(
            "start", torch.tensor(start, dtype=torch.get_default_dtype())
        )
        self.register_buffer(
            "cutoff", torch.tensor(cutoff, dtype=torch.get_default_dtype())
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        
        x = (x.clamp(self.start, self.cutoff) - self.start) / (self.cutoff - self.start)
        envelope = (
            1.0
            - ((self.p + 1.0) * (self.p + 2.0) / 2.0) * torch.pow(x, self.p)
            + self.p * (self.p + 2.0) * torch.pow(x, self.p + 1)
            - (self.p * (self.p + 1.0) / 2) * torch.pow(x, self.p + 2)
        )

        return envelope

    def __repr__(self):
        return f"{self.__class__.__name__}(p={self.p}, r_min={self.start}, r_max(cutoff)={self.cutoff})"

    @staticmethod
    def calculate_envelope(
        x: torch.Tensor, r_max: torch.Tensor, p: torch.Tensor
    ) -> torch.Tensor:
        r_over_r_max = x / r_max
        envelope = (
            1.0
            - ((p + 1.0) * (p + 2.0) / 2.0) * torch.pow(r_over_r_max, p)
            + p * (p + 2.0) * torch.pow(r_over_r_max, p + 1)
            - (p * (p + 1.0) / 2) * torch.pow(r_over_r_max, p + 2)
        )
        return envelope * (x < r_max)


class AgnesiTransform(torch.nn.Module):
    """
    From MACE https://github.com/ACEsuit/mace https://doi.org/10.1063/5.0158783
    """

    def __init__(
        self,
        q: float = 0.9183,
        p: float = 4.5791,
        a: float = 1.0805,
        trainable=False,
    ):
        super().__init__()
        self.register_buffer("q", torch.tensor(q, dtype=torch.get_default_dtype()))
        self.register_buffer("p", torch.tensor(p, dtype=torch.get_default_dtype()))
        self.register_buffer("a", torch.tensor(a, dtype=torch.get_default_dtype()))
        self.register_buffer(
            "covalent_radii",
            torch.tensor(
                ase.data.covalent_radii,
                dtype=torch.get_default_dtype(),
            ),
        )
        if trainable:
            self.a = torch.nn.Parameter(torch.tensor(1.0805, requires_grad=True))
            self.q = torch.nn.Parameter(torch.tensor(0.9183, requires_grad=True))
            self.p = torch.nn.Parameter(torch.tensor(4.5791, requires_grad=True))

    def forward(
        self,
        x: Tensor,
        node_attrs: Tensor,
        edge_index: Tensor,
        atomic_numbers: Tensor,
    ) -> Tensor:
        source = edge_index[0]
        target = edge_index[1]
        node_atomic_numbers = atomic_numbers[torch.argmax(node_attrs, dim=1)].unsqueeze(
            -1
        ) 
        Z_u = node_atomic_numbers[source]
        Z_v = node_atomic_numbers[target]
        r_0: torch.Tensor = 0.5 * (
            self.covalent_radii[Z_u] + self.covalent_radii[Z_v]
        ) 
        r_over_r_0 = x / r_0
        return (
            1
            + (
                self.a
                * torch.pow(r_over_r_0, self.q)
                / (1 + torch.pow(r_over_r_0, self.q - self.p))
            )
        ).reciprocal_()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(a={self.a:.4f}, q={self.q:.4f}, p={self.p:.4f})"
        )


class SoftTransform(torch.nn.Module):
    """
    From MACE https://github.com/ACEsuit/mace https://doi.org/10.1063/5.0158783
    """

    def __init__(self, a: float = 0.2, b: float = 3.0, trainable=False):
        super().__init__()
        self.register_buffer(
            "covalent_radii",
            torch.tensor(
                ase.data.covalent_radii,
                dtype=torch.get_default_dtype(),
            ),
        )
        if trainable:
            self.a = torch.nn.Parameter(torch.tensor(a, requires_grad=True))
            self.b = torch.nn.Parameter(torch.tensor(b, requires_grad=True))
        else:
            self.register_buffer("a", torch.tensor(a))
            self.register_buffer("b", torch.tensor(b))

    def forward(
        self,
        x: Tensor,
        node_attrs: Tensor,
        edge_index: Tensor,
        atomic_numbers: Tensor,
    ) -> Tensor:
        source = edge_index[0]
        target = edge_index[1]
        node_atomic_numbers = atomic_numbers[torch.argmax(node_attrs, dim=1)].unsqueeze(
            -1
        )
        Z_u = node_atomic_numbers[source]
        Z_v = node_atomic_numbers[target]
        r_0 = (self.covalent_radii[Z_u] + self.covalent_radii[Z_v]) / 4
        r_over_r_0 = x / r_0
        y = (
            x
            + (1 / 2) * torch.tanh(-r_over_r_0 - self.a * torch.pow(r_over_r_0, self.b))
            + 1 / 2
        )
        return y

    def __repr__(self):
        return f"{self.__class__.__name__}(a={self.a.item()}, b={self.b.item()})"


class ChebychevBasis(torch.nn.Module):
    """ Not recommended for use in TACE """
    def __init__(self, cutoff: float, num_basis=8):
        super().__init__()
        self.register_buffer(
            "n",
            torch.arange(1, num_basis + 1, dtype=torch.get_default_dtype()).unsqueeze(
                0
            ),
        )
        self.num_basis = num_basis
        self.cutoff = cutoff

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.repeat(1, self.num_basis)
        n = self.n.repeat(len(x), 1)
        return torch.special.chebyshev_polynomial_t(x, n)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(cutoff={self.cutoff}, num_basis={self.num_basis},"
        )


class GaussianBasis(torch.nn.Module):
    """ Not recommended for use in TACE """
    def __init__(self, cutoff: float, num_basis=128, trainable=False):
        super().__init__()
        gaussian_weights = torch.linspace(
            start=0.0, end=cutoff, steps=num_basis, dtype=torch.get_default_dtype()
        )
        if trainable:
            self.gaussian_weights = torch.nn.Parameter(
                gaussian_weights, requires_grad=True
            )
        else:
            self.register_buffer("gaussian_weights", gaussian_weights)
        self.coeff = -0.5 / (cutoff / (num_basis - 1)) ** 2

        self.num_basis = num_basis
        self.cutoff = cutoff

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # [..., 1]
        x = x - self.gaussian_weights
        return torch.exp(self.coeff * torch.pow(x, 2))

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(cutoff={self.cutoff}, num_basis={self.num_basis},"
        )


class ZBLBasis(torch.nn.Module):
    """
    Metal units
    From LAMMPS pair_zbl_const.h, code from MACE
    Implementation of the Ziegler-Biersack-Littmark (ZBL) potential
    with a polynomial cutoff envelope.
    """

    p: torch.Tensor

    def __init__(self, p:int = 6, trainable: bool = False, **kwargs):
        super().__init__()
        # Pre-calculate the p coefficients for the ZBL potential
        self.register_buffer(
            "c",
            torch.tensor(
                [0.1818, 0.5099, 0.2802, 0.02817], dtype=torch.get_default_dtype()
            ),
        )
        self.register_buffer("p", torch.tensor(p, dtype=torch.int))
        self.register_buffer(
            "covalent_radii",
            torch.tensor(
                ase.data.covalent_radii,
                dtype=torch.get_default_dtype(),
            ),
        )
        if trainable:
            self.a_exp = torch.nn.Parameter(torch.tensor(0.300, requires_grad=True))
            self.a_prefactor = torch.nn.Parameter(
                torch.tensor(0.4543, requires_grad=True)
            )
        else:
            self.register_buffer("a_exp", torch.tensor(0.300))
            self.register_buffer("a_prefactor", torch.tensor(0.4543))

    def forward(
        self,
        x: torch.Tensor,
        node_attrs: torch.Tensor,
        edge_index: torch.Tensor,
        atomic_numbers: torch.Tensor,
    ) -> torch.Tensor:
        source = edge_index[0]
        target = edge_index[1]
        node_atomic_numbers = atomic_numbers[torch.argmax(node_attrs, dim=1)].unsqueeze(-1)
        Z_u = node_atomic_numbers[source]
        Z_v = node_atomic_numbers[target]
        a = (
            self.a_prefactor
            * 0.529
            / (torch.pow(Z_u, self.a_exp) + torch.pow(Z_v, self.a_exp))
        )
        r_over_a = x / a
        phi = (
            self.c[0] * torch.exp(-3.2 * r_over_a)
            + self.c[1] * torch.exp(-0.9423 * r_over_a)
            + self.c[2] * torch.exp(-0.4028 * r_over_a)
            + self.c[3] * torch.exp(-0.2016 * r_over_a)
        )
        v_edges = (14.3996 * Z_u * Z_v) / x * phi
        r_max = self.covalent_radii[Z_u] + self.covalent_radii[Z_v]
        envelope = PolynomialCutoff.calculate_envelope(x, r_max, self.p)
        v_edges = 0.5 * v_edges * envelope
        V_ZBL = scatter_sum(v_edges, target, dim=0, dim_size=node_attrs.size(0))
        return V_ZBL.squeeze(-1)

    def __repr__(self):
        return f"{self.__class__.__name__}(c={[float(f'{c:.4f}') for c in self.c.tolist()]})"


class RadialBasis(torch.nn.Module):
    def __init__(
        self,
        cutoff: float = 6.0,
        num_basis: int = 8,
        polynomial_cutoff: int = 5.0,
        radial_basis: str = "bessel",
        distance_transform=None,
        order: int | List[int] = [0],
        trainable: bool = False,
        apply_cutoff: bool = True,
    ):
        super().__init__()
        if radial_basis == "bessel" or radial_basis == "j0":
            self.radial_fn = j0_SphericalBesselBasis(
                cutoff=cutoff, num_basis=num_basis, trainable=trainable
            )
        elif radial_basis == "jn":
            self.radial_fn = jn_SphericalBesselBasis(
                cutoff=cutoff,
                order=order,
                num_basis=num_basis,
                trainable=trainable,
            )
        elif radial_basis == "normalized_bessel" or radial_basis == "n_j0":
            self.radial_fn = Normalized_j0_SphericalBesselBasis(
                r_max=cutoff,
                original_basis_kwargs={"cutoff": cutoff, "num_basis": num_basis},
            )
        elif radial_basis == "chebychev":
            self.radial_fn = ChebychevBasis(
                cutoff=cutoff,
                num_basis=num_basis,
            )
        elif radial_basis == "gaussian":
            self.radial_fn = GaussianBasis(
                cutoff=cutoff,
                num_basis=num_basis,
                trainable=trainable,
            )
        else:
            raise ValueError(f"Unknown radial_basis: {radial_basis}")
        if distance_transform == "Agnesi":
            self.distance_transform = AgnesiTransform()
        elif distance_transform == "Soft":
            self.distance_transform = SoftTransform()
        self.cutoff_fn = PolynomialCutoff(cutoff=cutoff, p=polynomial_cutoff)

        if not isinstance(num_basis, int):
            num_basis = sum(num_basis)
            self.out_dim = num_basis
            self.num_basis = num_basis
        else:
            self.out_dim = num_basis
            self.num_basis = num_basis

        self.apply_cutoff = apply_cutoff
        self.for_copy = {
            "cutoff": cutoff,
            "num_basis": num_basis,
            "polynomial_cutoff": polynomial_cutoff,
            "radial_basis": radial_basis,
            "distance_transform": distance_transform,
            "order": order,
            "trainable": trainable,
            "apply_cutoff": apply_cutoff,
        }

    def forward(
        self,
        edge_length: torch.Tensor,
        node_attrs: torch.Tensor,
        edge_index: torch.Tensor,
        atomic_numbers: torch.Tensor,
    ) -> torch.Tensor:
        cutoff = self.cutoff_fn(edge_length)
        if hasattr(self, "distance_transform"):
            edge_length = self.distance_transform(
                edge_length, node_attrs, edge_index, atomic_numbers
            )
        radial = self.radial_fn(edge_length)
        if self.apply_cutoff:
            return radial * cutoff, None
        else:
            return radial, cutoff

    # j1 j2 j3 ...
    def copy(self, lmax: int = -1):
        return self.replicate()

    def replicate(self):
        for_copy = self.for_copy
        return self.__class__(**for_copy)
    
# # draw
# import torch
# import matplotlib.pyplot as plt
# import numpy as np

# torch.set_default_dtype(torch.float64)
# cutoff = 6.0
# num_basis = 8
# num_basis_list = [15] # [8, 9, 10]

# j0 = j0_SphericalBesselBasis(cutoff=cutoff, num_basis=num_basis)
# jn = jn_SphericalBesselBasis(cutoff=cutoff, order=8, num_basis=num_basis_list)

# x = torch.linspace(0, cutoff, 200).unsqueeze(-1)

# a = j0(x)  # shape: [200, 8]
# b = jn(x)  # shape: [200, 8]


# x_np = x.squeeze().numpy()        # shape: [200]
# a = a.detach().numpy()         # shape: [200, 8]
# b = b.detach().numpy()         # shape: [200, 8]


# plt.figure(figsize=(10, 5))

# for i in range(a.shape[1]):
#     # plt.plot(x_np, a[:, i], label=f'j0_basis_{i}', linestyle='-')
#     plt.plot(x_np, b[:, i], label=f'jn_basis_{i}', linestyle='--')

# plt.xlabel("x")
# plt.ylabel("Basis value")
# plt.title("Comparison of j0 vs jn spherical Bessel basis")
# plt.legend(fontsize='small', ncol=2)
# plt.grid(True)
# plt.tight_layout()
# plt.show()


################################################################################
# Authors: Zemin Xu
# License: Academic / Non-Commercial Use Only
################################################################################

import pickle
import contextlib
import logging
from typing import Dict, List
import yaml
from pathlib import Path
from packaging import version

from torch import Tensor
from omegaconf import DictConfig, ListConfig


import numpy as np
import torch


def set_global_seed(cfg: Dict) -> None:
    seed = cfg["misc"].get("global_seed", 42)
    split_seed = cfg["dataset"].get("split_seed", 42)
    # torch.backends.cudnn.deterministic = True
    # torch.backends.cudnn.benchmark = False
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    logging.info(f"Global seed: {seed}")
    logging.info(f"Split seed: {split_seed}")


def set_precision(cfg: Dict) -> None:
    precision = cfg["trainer"]["precision"]
    FLOAT64 = {"64-true", "64", 64}
    FLOAT32 = {"32-true", "32", 32}
    FLOAT16 = {"16-mixed", "16", 16}
    BFLOAT16 = {"bf16-mixed", "bf16"}
    ALLOWED_PRECISIONS = FLOAT64 | FLOAT32 | FLOAT16 | BFLOAT16
    try:
        assert precision is not None and precision in ALLOWED_PRECISIONS, (
            f"Invalid precision setting: {precision!r}. "
            f"Must be one of: {ALLOWED_PRECISIONS}"
        )
    except Exception as e:
        raise RuntimeError(f"The cfg.trainer.precision value must be specified.") from e

    if precision in FLOAT64:
        torch.set_default_dtype(torch.float64)
    elif precision in FLOAT32:
        torch.set_default_dtype(torch.float32)
    elif precision in FLOAT16 or precision in BFLOAT16:
        torch.set_default_dtype(torch.float32)

    # === allow_tf32 ===
    allow_tf32 = cfg["misc"].get("allow_tf32", False)
    if torch.cuda.is_available():
        if allow_tf32:
            torch.backends.cuda.matmul.allow_tf32 = allow_tf32
            torch.backends.cudnn.allow_tf32 = allow_tf32

# torch 2.9
# def set_precision(cfg: Dict) -> None:
#     # === init ===
#     global _GLOBAL_STATE_INITIALIZED
#     if not _GLOBAL_STATE_INITIALIZED:
#         torch.set_default_dtype(torch.float64)
#         if torch.cuda.is_available():
#             if _TORCH_GE_2_9:
#                 torch.backends.fp32_precision = "ieee"
#             else:
#                 torch.backends.cuda.matmul.allow_tf32 = False
#                 torch.backends.cudnn.allow_tf32 = False
#         _GLOBAL_STATE_INITIALIZED = True

#     # === set tensor dtype ===
#     precision = cfg["trainer"]["precision"]
#     FLOAT64 = {"64-true", "64", 64}
#     FLOAT32 = {"32-true", "32", 32}
#     FLOAT16 = {"16-mixed", "16", 16}
#     BFLOAT16 = {"bf16-mixed", "bf16"}
#     ALLOWED_PRECISIONS = FLOAT64 | FLOAT32 | FLOAT16 | BFLOAT16
#     try:
#         assert precision is not None and precision in ALLOWED_PRECISIONS, (
#             f"Invalid precision setting: {precision!r}. "
#             f"Must be one of: {ALLOWED_PRECISIONS}"
#         )
#     except Exception as e:
#         raise RuntimeError(f"The cfg.trainer.precision value must be specified.") from e

#     if precision in FLOAT32:
#         torch.set_default_dtype(torch.float32)
#     elif precision in FLOAT16 or precision in BFLOAT16:
#         torch.set_default_dtype(torch.float32)

#     # === allow_tf32 ===
#     allow_tf32 = cfg["misc"].get("allow_tf32", False)
#     if torch.cuda.is_available():
#         if _TORCH_GE_2_9:
#             current_precision = torch.backends.fp32_precision
#             desired_precision = "tf32" if allow_tf32 else "ieee"
#             if current_precision != desired_precision:
#                 torch.backends.fp32_precision = desired_precision
#         else:
#             if torch.backends.cuda.matmul.allow_tf32 is not allow_tf32:
#                 torch.backends.cuda.matmul.allow_tf32 = allow_tf32
#                 torch.backends.cudnn.allow_tf32 = allow_tf32
    
def num_params(model) -> None:
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def log_parameters(model) -> None:
    logging.debug(f"Total number of parameters in the model: {num_params(model)}")
    for name, param in model.named_parameters():
        if param.requires_grad:
            logging.debug(f"Layer: {name}, Number of parameters: {param.numel()}")


def to_serializable(obj):
    if isinstance(obj, (int, float, str, bool)) or obj is None:
        return obj
    elif isinstance(obj, (list, tuple)):
        return [to_serializable(i) for i in obj]
    elif isinstance(obj, dict):
        return {k: to_serializable(v) for k, v in obj.items()}
    elif hasattr(obj, "__dict__"):
        return {
            k: to_serializable(v)
            for k, v in vars(obj).items()
            # if not k.startswith("_")
        }
    else:
        return str(obj)


def log_statistics_to_yaml(obj) -> None:
    # def float_representer(dumper, value):
    #     return dumper.represent_scalar('tag:yaml.org,2002:float', f'{value:.15f}')
    # yaml.add_representer(float, float_representer)
    if obj is not None:
        for idx, stat in enumerate(obj):
            filename = Path(".") / f"statistics_{idx}.yaml"
            with open(filename, "w") as f:
                # yaml.safe_dump(obj, f, sort_keys=False, allow_unicode=True)
                yaml.dump(stat, f, sort_keys=False, allow_unicode=True)


@contextlib.contextmanager
def torch_default_dtype(dtype):
    default_dtype = torch.get_default_dtype()
    try:
        torch.set_default_dtype(dtype)
        yield
    finally:
        torch.set_default_dtype(default_dtype)


def is_rank_0():
    if not torch.distributed.is_available():
        return True
    if not torch.distributed.is_initialized():
        return True
    return torch.distributed.get_rank() == 0


def save_full_cfg(cfg: Dict):
    if is_rank_0():
        path = Path(".") / "_tace.yaml"
        with open(f"{path}", "w") as f:
            yaml.dump(cfg, f, sort_keys=False)


def deep_convert(cfg):
    if isinstance(cfg, DictConfig):
        cfg = dict(cfg)
        for key, value in cfg.items():
            cfg[key] = deep_convert(value)
        return cfg
    elif isinstance(cfg, ListConfig):
        cfg = list(cfg)
        for i, item in enumerate(cfg):
            cfg[i] = deep_convert(item)
        return cfg
    else:
        return cfg


def voigt_to_matrix(t: Tensor, **kwargs):
    """
    Convert voigt notation to matrix notation
    """
    if t.shape == (3, 3):
        return t
    if t.shape == (6,):
        return torch.tensor(
            [
                [t[0], t[5], t[4]],
                [t[5], t[1], t[3]],
                [t[4], t[3], t[2]],
            ],
            dtype=t.dtype,
        )
    if t.shape == (9,):
        return t.view(3, 3)

    raise ValueError(
        f"Stress tensor must be of shape (6,) or (3, 3), or (9,) but has shape {t.shape}"
    )



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import torch


ACT = {
    None: torch.nn.Identity,
    "None": torch.nn.Identity,
    "null": torch.nn.Identity,
    "identity": torch.nn.Identity,
    "relu": torch.nn.ReLU,
    "leaky_relu": torch.nn.LeakyReLU,
    "prelu": torch.nn.PReLU,
    "elu": torch.nn.ELU,
    "selu": torch.nn.SELU,
    "gelu": torch.nn.GELU,
    "silu": torch.nn.SiLU,  # Swish
    "mish": torch.nn.Mish,
    "softplus": torch.nn.Softplus,
    "softsign": torch.nn.Softsign,
    "tanh": torch.nn.Tanh,
    "sigmoid": torch.nn.Sigmoid,
    "hardtanh": torch.nn.Hardtanh,
    "hardswish": torch.nn.Hardswish,
    "hardsigmoid": torch.nn.Hardsigmoid,
    "tanhshrink": torch.nn.Tanhshrink,
}
"""Mapping from activation function name to torch.nn modules.

You can specify the key to use these activation functions in TACE.

- None: ``torch.nn.Identity``
- identity: ``torch.nn.Identity``
- relu: ``torch.nn.ReLU``
- leaky_relu: ``torch.nn.LeakyReLU``
- prelu: ``torch.nn.PReLU``
- elu: ``torch.nn.ELU``
- selu: ``torch.nn.SELU``
- gelu: ``torch.nn.GELU``
- silu: ``torch.nn.SiLU`` (Swish)
- mish: ``torch.nn.Mish``
- softplus: ``torch.nn.Softplus``
- softsign: ``torch.nn.Softsign``
- tanh: ``torch.nn.Tanh``
- sigmoid: ``torch.nn.Sigmoid``
- hardtanh: ``torch.nn.Hardtanh``
- hardswish: ``torch.nn.Hardswish``
- hardsigmoid: ``torch.nn.Hardsigmoid``
- tanhshrink: ``torch.nn.Tanhshrink``
"""



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Dict, List


import torch
from torch import Tensor
from torchmetrics import Metric


from ..dataset.quantity import (
    MAE_PROPERTY,
    RMSE_PROPERTY,
    MAE_PER_ATOM_PROPERTY,
    RMSE_PER_ATOM_PROPERTY,
)

SCALE = 1000.0  # for example, metric units from eV to meV


def expand_dims_to(T: torch.Tensor, n_dim: int, dim: int = -1) -> torch.Tensor:
    '''jit-safe'''
    while T.ndim < n_dim:
        T = T.unsqueeze(dim)
    return T


class MAE(Metric):
    def __init__(self, scale: float = SCALE):
        super().__init__()
        self.scale = scale
        self.add_state("sum_abs_error", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("count", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: torch.Tensor, targets: torch.Tensor):
        abs_error = torch.abs(preds - targets)
        self.sum_abs_error += abs_error.sum()
        self.count += targets.numel()

    def compute(self):
        return (self.sum_abs_error / self.count) * self.scale


class RMSE(Metric):
    def __init__(self, scale: float = SCALE):
        super().__init__()
        self.scale = scale
        self.add_state(
            "sum_squared_error", default=torch.tensor(0.0), dist_reduce_fx="sum"
        )
        self.add_state("count", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: Tensor, targets: Tensor):
        squared_error = (preds - targets) ** 2
        self.sum_squared_error += squared_error.sum()
        self.count += targets.numel()

    def compute(self):
        return torch.sqrt(self.sum_squared_error / self.count) * self.scale


class PerAtomMAE(MAE):
    def update(self, preds: Tensor, targets: Tensor, ptr: Tensor):
        num_nodes = ptr[1:] - ptr[:-1]  # [B]
        num_nodes = expand_dims_to(num_nodes, preds.ndim, dim=-1)
        preds_per_atom = preds / num_nodes
        targets_per_atom = targets / num_nodes
        super().update(preds_per_atom, targets_per_atom)


class PerAtomRMSE(RMSE):
    def update(self, preds: Tensor, targets: Tensor, ptr: Tensor):
        num_nodes = ptr[1:] - ptr[:-1]  # [B]
        num_nodes = expand_dims_to(num_nodes, preds.ndim, dim=-1)
        preds_per_atom = preds / num_nodes
        targets_per_atom = targets / num_nodes
        super().update(preds_per_atom, targets_per_atom)


class PolarizationMetric(Metric):
    def __init__(
        self, metric_type: str = "mae", per_atom: bool = False, scale: float = SCALE
    ):
        """
        Args:
            metric_type: "mae" or "rmse"
            per_atom: whether to normalize by number of atoms
            scale: scaling factor
        """
        super().__init__()
        assert metric_type in ["mae", "rmse"], "metric_type must be 'mae' or 'rmse'"
        self.metric_type = metric_type
        self.per_atom = per_atom
        self.scale = scale

        if self.metric_type == "mae":
            self.add_state(
                "sum_abs_error", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )
        else:  # rmse
            self.add_state(
                "sum_squared_error", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )

        self.add_state("count", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: Tensor, targets: Tensor, label: Dict[str, Tensor]):
        lattice = label["lattice"]
        error = preds - targets

        # Project to lattice basis
        error = torch.einsum("bi, bij -> bj", error, torch.linalg.inv(lattice))
        error = torch.remainder(error, 1.0)
        error = torch.where(error > 0.5, error - 1.0, error)
        error = torch.where(error < -0.5, error + 1.0, error)
        error = torch.einsum("bi, bij -> bj", error, lattice)

        # Normalize per atom if needed
        if self.per_atom:
            ptr = label["ptr"]
            num_atoms = (ptr[1:] - ptr[:-1]).reshape(-1, 1)
            error = error / num_atoms

        if self.metric_type == "mae":
            abs_error = torch.abs(error)
            self.sum_abs_error += abs_error.sum()
        else:  # rmse
            squared_error = error**2
            self.sum_squared_error += squared_error.sum()

        self.count += targets.numel()

    def compute(self):
        if self.metric_type == "mae":
            return self.sum_abs_error / self.count * self.scale
        else:  # rmse
            return torch.sqrt(self.sum_squared_error / self.count) * self.scale

class FinalCollinearMagmomsMetric(Metric):
    def __init__(self, metric_type: str = "mae", scale: float = SCALE):
        """
        Args:
            metric_type: "mae" or "rmse"
            scale: scaling factor
        """
        super().__init__()
        assert metric_type in ["mae", "rmse"], "metric_type must be 'mae' or 'rmse'"
        self.metric_type = metric_type
        self.scale = scale

        if self.metric_type == "mae":
            self.add_state(
                "sum_abs_error", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )
        else:  # rmse
            self.add_state(
                "sum_squared_error", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )

        self.add_state("count", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: Tensor, targets: Tensor):
        error = abs(preds) - abs(targets)
        if self.metric_type == "mae":
            abs_error = torch.abs(error)
            self.sum_abs_error += abs_error.sum()
        else:  # rmse
            squared_error = error**2
            self.sum_squared_error += squared_error.sum()
        self.count += targets.numel()

    def compute(self):
        if self.metric_type == "mae":
            return self.sum_abs_error / self.count * self.scale
        else:  # rmse
            return torch.sqrt(self.sum_squared_error / self.count) * self.scale
        
class PartialHessiansMetric(Metric):
    def __init__(self, metric_type: str = "mae", scale: float = SCALE):
        """
        Args:
            metric_type: "mae" or "rmse"
            scale: scaling factor
        """
        super().__init__()
        assert metric_type in ["mae", "rmse"], "metric_type must be 'mae' or 'rmse'"
        self.metric_type = metric_type
        self.scale = scale

        if self.metric_type == "mae":
            self.add_state(
                "sum_abs_error", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )
        else:  # rmse
            self.add_state(
                "sum_squared_error", default=torch.tensor(0.0), dist_reduce_fx="sum"
            )

        self.add_state("count", default=torch.tensor(0), dist_reduce_fx="sum")


    def update(self, preds: Tensor, targets: Tensor):
        error = abs(preds) - abs(targets)
        if self.metric_type == "mae":
            abs_error = torch.abs(error)
            self.sum_abs_error += abs_error.sum()
        else:  # rmse
            squared_error = error**2
            self.sum_squared_error += squared_error.sum()
        self.count += targets.numel()

    def update(self, pred: Dict[str, torch.Tensor | List[torch.Tensor]], label: Dict[str, torch.Tensor]):

        true_hessian_flat = label["hessians"]
        num_atoms_per_graph = label["ptr"][1:] - label["ptr"][:-1]
        jacs_per_graph = pred["jacs_per_graph"]
        samples_per_graph = pred["samples_per_graph"]

        offset = 0
        error_list = []
        for jac_pred, samples, n_g in zip(jacs_per_graph, samples_per_graph, num_atoms_per_graph):
            hess_size = n_g * 3 * n_g * 3
            hess_flat_g = true_hessian_flat[offset : offset + hess_size]
            hess_true = hess_flat_g.reshape(n_g, 3, n_g, 3)
            offset += hess_size

            atom_idx = samples[:, 0]
            xyz_idx = samples[:, 1]
            jac_true = hess_true[atom_idx, xyz_idx]
            error = torch.square((jac_pred - jac_true)).view(-1)
            error_list.append(error)
        errors = torch.cat(error_list, dim=-1)

        if self.metric_type == "mae":
            abs_error = torch.abs(errors)
            self.sum_abs_error += abs_error.sum()
        else:  # rmse
            squared_error = errors**2
            self.sum_squared_error += squared_error.sum()
        self.count += errors.numel() 
   
    def compute(self):
        if self.metric_type == "mae":
            return self.sum_abs_error / self.count * self.scale
        else:  # rmse
            return torch.sqrt(self.sum_squared_error / self.count) * self.scale
        
def build_metrics(prefix: str, loss_property: List[str]) -> Dict[str, Metric]:
    metrics = {}

    def add_metrics(property_name):
        if property_name in MAE_PROPERTY:
            metrics[f"{prefix}/{property_name}_mae"] = MAE()
        if property_name in RMSE_PROPERTY:
            metrics[f"{prefix}/{property_name}_rmse"] = RMSE()
        if property_name in MAE_PER_ATOM_PROPERTY:
            metrics[f"{prefix}/{property_name}_per_atom_mae"] = PerAtomMAE()
        if property_name in RMSE_PER_ATOM_PROPERTY:
            metrics[f"{prefix}/{property_name}_per_atom_rmse"] = PerAtomRMSE()
        if property_name == "polarization":
            metrics[f"{prefix}/{property_name}_mae"] = PolarizationMetric("mae", False)
            metrics[f"{prefix}/{property_name}_rmse"] = PolarizationMetric("rmse", False)
            metrics[f"{prefix}/{property_name}_per_atom_mae"] = PolarizationMetric("mae", True)
            metrics[f"{prefix}/{property_name}_per_atom_rmse"] = PolarizationMetric("rmse", True)
        # if property_name == "final_collinear_magmoms":
        #     metrics[f"{prefix}/{property_name}_mae"] = FinalCollinearMagmomsMetric("mae")
        #     metrics[f"{prefix}/{property_name}_rmse"] = FinalCollinearMagmomsMetric("rmse")
        if property_name == "hessians":
            metrics[f"{prefix}/{property_name}_mae"] = PartialHessiansMetric("mae")
            metrics[f"{prefix}/{property_name}_rmse"] = PartialHessiansMetric("rmse")

    for p in loss_property:
        add_metrics(p)

    return metrics


def update_metrics(metrics, prefix, pred, label, loss_property):
    for p in loss_property:

        if p not in pred:
            continue

        ptr = label["ptr"]
        output_value = pred[p]
        batch_value = label[p]
        if p in MAE_PROPERTY:
            metrics[f"{prefix}/{p}_mae"](output_value, batch_value)
        if p in RMSE_PROPERTY:
            metrics[f"{prefix}/{p}_rmse"](output_value, batch_value)
        if p in MAE_PER_ATOM_PROPERTY:
            metrics[f"{prefix}/{p}_per_atom_mae"](output_value, batch_value, ptr)
        if p in RMSE_PER_ATOM_PROPERTY:
            metrics[f"{prefix}/{p}_per_atom_rmse"](output_value, batch_value, ptr)
        if p == "polarization":
            metrics[f"{prefix}/{p}_mae"](output_value, batch_value, label)
            metrics[f"{prefix}/{p}_rmse"](output_value, batch_value, label)
            metrics[f"{prefix}/{p}_per_atom_mae"](output_value, batch_value, label)
            metrics[f"{prefix}/{p}_per_atom_rmse"](output_value, batch_value, label)
        # if p == "final_collinear_magmoms":
        #     metrics[f"{prefix}/{p}_mae"](output_value, batch_value)
        #     metrics[f"{prefix}/{p}_rmse"](output_value, batch_value)
        if p == "hessians":
            metrics[f"{prefix}/{p}_mae"](pred, label)
            metrics[f"{prefix}/{p}_rmse"](pred, label)



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import logging
from typing import Dict, Union

import torch


from ..models.v1.mlp import LinearLayer, LoRALinearLayer
from ..models.v1.linear import (
    Linear, 
    ElementLinear, 
    CWLinear, 
    ElementCWLinear,
    LoRALinear,
    LoRAElementLinear,
    LoRACWLinear,
    LoRAElementCWLinear,
)

def to_lora_mlp(
    linear: LinearLayer,
    lora_config: Dict[str, int | float | bool],
) -> LoRALinearLayer:
    
    arguments = {
        'in_dim': linear.in_dim,
        'out_dim': linear.out_dim,
        'alpha': linear.alpha,
        'bias': linear.bias is not None,
        'lora_r': int(lora_config.get('r', -1)),
        'lora_alpha': float(lora_config.get('alpha', lora_config.get('r', -1))),
    }

    if arguments['lora_r'] <= 0:
        return linear
    
    lora_linear = LoRALinearLayer(**arguments)
    lora_linear = lora_linear.to(
        device=linear.weight.device,
        dtype=linear.weight.dtype,
    )
    with torch.no_grad():
        lora_linear.weight.copy_(linear.weight)
        if arguments['bias']:
            lora_linear.bias.copy_(linear.bias)
 
    return lora_linear

def from_lora_mlp(lora_linear: LoRALinearLayer) -> LinearLayer:
    arguments = {
        'in_dim': lora_linear.in_dim,
        'out_dim': lora_linear.out_dim,
        'alpha': lora_linear.alpha,
        'bias': lora_linear.bias is not None,
    }
    linear = LinearLayer(**arguments)
    linear = linear.to(
        device=lora_linear.weight.device,
        dtype=lora_linear.weight.dtype,
    )
    with torch.no_grad():
        delta_w = lora_linear.lora_A @ (lora_linear.lora_B * lora_linear.scaling)
        linear.weight.copy_(lora_linear.weight + delta_w)
        if arguments['bias']:
            linear.bias.copy_(lora_linear.bias)
    return linear

def to_lora_linear(
    linear: Union[Linear, ElementLinear, CWLinear, ElementCWLinear],
    lora_config: Dict[str, int | float | bool],
) -> Union[LoRALinear, LoRAElementLinear, LoRACWLinear, LoRAElementCWLinear]:
    
    arguments = {
        'in_dim': linear.in_dim,
        'out_dim': linear.out_dim,
        'bias': linear.bias is not None,
        'l': linear.l,
        'lora_r': int(lora_config.get('r', 4)),
        'lora_alpha': float(lora_config.get('alpha', lora_config.get('r', 4))),
        'element_aware': bool(lora_config.get('element_aware', True)),
        'atomic_numbers': getattr(linear, 'atomic_numbers', None)
    }

    if arguments['lora_r'] <= 0:
        return linear
    
    if isinstance(linear, Linear):
        lora_linear = LoRALinear(**arguments)
        lora_linear = lora_linear.to(
            device=linear.weight.device,
            dtype=linear.weight.dtype,
        )
        with torch.no_grad():
            lora_linear.weight.copy_(linear.weight)
            if arguments['bias']:
                lora_linear.bias.copy_(linear.bias)
    elif isinstance(linear, CWLinear):
        lora_linear = LoRACWLinear(**arguments)
        lora_linear = lora_linear.to(
            device=linear.weight.device,
            dtype=linear.weight.dtype,
        )
        with torch.no_grad():
            lora_linear.weight.copy_(linear.weight)
            if arguments['bias']:
                lora_linear.bias.copy_(linear.bias)
    elif isinstance(linear, ElementLinear):
        lora_linear = LoRAElementLinear(**arguments)
        lora_linear = lora_linear.to(
            device=linear.weights.device,
            dtype=linear.weights.dtype,
        )
        with torch.no_grad():
            lora_linear.weights.copy_(linear.weights)
            if arguments['bias']:
                lora_linear.bias.copy_(linear.bias)
    elif isinstance(linear, ElementCWLinear):
        lora_linear = LoRAElementCWLinear(**arguments)
        lora_linear = lora_linear.to(
            device=linear.weights.device,
            dtype=linear.weights.dtype,
        )
        with torch.no_grad():
            lora_linear.weights.copy_(linear.weights)
            if arguments['bias']:
                lora_linear.bias.copy_(linear.bias)

    return lora_linear

def from_lora_linear(
    lora_linear: Union[
        LoRALinear,
        LoRAElementLinear,
        LoRACWLinear,
        LoRAElementCWLinear,
    ]
) -> Union[
    Linear,
    ElementLinear,
    CWLinear,
    ElementCWLinear,
]:
    #  === Linear  ===
    if isinstance(lora_linear, LoRALinear):
        arguments = {
            'in_dim': lora_linear.in_dim,
            'out_dim': lora_linear.out_dim,
            'bias': lora_linear.bias is not None,
            'l': lora_linear.l,
        }
        linear = Linear(**arguments).to(
            device=lora_linear.weight.device,
            dtype=lora_linear.weight.dtype,
        )
        with torch.no_grad():
            delta_w = lora_linear.lora_A @ (
                lora_linear.lora_B * lora_linear.scaling
            )
            linear.weight.copy_(lora_linear.weight + delta_w)
            if arguments['bias']:
                linear.bias.copy_(lora_linear.bias)
        return linear

    # === CWLinear  ===
    elif isinstance(lora_linear, LoRACWLinear):
        arguments = {
            'in_dim': lora_linear.in_dim,
            'out_dim': lora_linear.out_dim,
            'bias': lora_linear.bias is not None,
            'l': lora_linear.l,
        }
        linear = CWLinear(**arguments).to(
            device=lora_linear.weight.device,
            dtype=lora_linear.weight.dtype,
        )
        with torch.no_grad():
            delta_w = lora_linear.lora_A @ (
                lora_linear.lora_B * lora_linear.scaling
            )
            linear.weight.copy_(lora_linear.weight + delta_w)
            if arguments['bias']:
                linear.bias.copy_(lora_linear.bias)
        return linear

    #  === ElementLinear ===
    elif isinstance(lora_linear, LoRAElementLinear):
        arguments = {
            'in_dim': lora_linear.in_dim,
            'out_dim': lora_linear.out_dim,
            'bias': lora_linear.bias is not None,
            'l': lora_linear.l,
            'atomic_numbers': lora_linear.atomic_numbers,
        }
        linear = ElementLinear(**arguments).to(
            device=lora_linear.weights.device,
            dtype=lora_linear.weights.dtype,
        )
        with torch.no_grad():
            if lora_linear.element_aware:
                delta_w = torch.einsum(
                    'zri, zor -> zoi',
                    lora_linear.lora_A,
                    lora_linear.lora_B * lora_linear.scaling,
                )
            else:
                delta_w = (lora_linear.lora_B * lora_linear.scaling) @ lora_linear.lora_A 
                delta_w = delta_w.unsqueeze(0)
            linear.weights.copy_(lora_linear.weights + delta_w)
            if arguments['bias']:
                linear.bias.copy_(lora_linear.bias)
        return linear

    # === ElementCWLinear ===
    elif isinstance(lora_linear, LoRAElementCWLinear):
        arguments = {
            'in_dim': lora_linear.in_dim,
            'out_dim': lora_linear.out_dim,
            'bias': lora_linear.bias is not None,
            'l': lora_linear.l,
            'atomic_numbers': lora_linear.atomic_numbers,
        }
        linear = ElementCWLinear(**arguments).to(
            device=lora_linear.weights.device,
            dtype=lora_linear.weights.dtype,
        )
        with torch.no_grad():
            delta_w = (lora_linear.lora_B.T @ lora_linear.lora_A.T) * lora_linear.scaling
            linear.weights.copy_(lora_linear.weights + delta_w.unsqueeze(0))
            if arguments['bias']:
                linear.bias.copy_(lora_linear.bias)
        return linear

    else:
        raise TypeError(f"Unsupported LoRA linear type: {type(lora_linear)}")


def inject_lora_into_model(
    model: torch.nn.Module,
    lora_configs: Dict[str, Dict[str, int | float]],
):
    to_replace = []
    for full_name, module in model.named_modules():
        if isinstance(module, LinearLayer):
            to_replace.append((full_name, module, to_lora_mlp))
        elif isinstance(module, Linear | ElementLinear | CWLinear | ElementCWLinear):
            to_replace.append((full_name, module, to_lora_linear))


    for full_name, module, replace_fn in to_replace:
        parent = model
        *path, attr = full_name.split(".")
        for p in path:
            parent = getattr(parent, p)

        if '.node_embedding.' in full_name:
            cfg = lora_configs['element_embedding']
        elif '.radial_net.' in full_name:
            cfg = lora_configs['radial_mlp']
        elif '.interactions.' in full_name:
            cfg = lora_configs['interaction']
        elif '.products.' in full_name:
            cfg = lora_configs['product']
        elif 'readout' in full_name:
            cfg = lora_configs['readout']
        else:
            raise

        setattr(
            parent,
            attr,
            replace_fn(module,  cfg)
        )

def to_lora_model(finetune_cfg: Dict, model: torch.nn.Module) -> torch.nn.Module:
    if not finetune_cfg: 
        return model

    # === LoRA ===
    lora = finetune_cfg.get('lora', {})
    if len(lora) > 0:
        inject_lora_into_model(model, lora)

    # === Freeze ===
    freeze = finetune_cfg.get('freeze', {})
    if len(freeze) > 0:
        name_to_param = dict(model.named_parameters())
        for name, flag in freeze.items():
            if name not in name_to_param:
                logging.warning(f"Parameter '{name}' not found in model")
                continue
            param = name_to_param[name]
            param.requires_grad = not bool(flag)
    logging.info(model)

    return model

def from_lora_to_merged_model(model: torch.nn.Module) -> torch.nn.Module:
    def _replace(module: torch.nn.Module):
        for name, child in module.named_children():
            # === LoRA MLP ===
            if isinstance(child, LoRALinearLayer):
                new_module = from_lora_mlp(child)
                setattr(module, name, new_module)
            # === LoRA Linear ===
            elif isinstance(
                child,
                (
                    LoRALinear,
                    LoRAElementLinear,
                    LoRACWLinear,
                    LoRAElementCWLinear,
                ),
            ):
                new_module = from_lora_linear(child)
                setattr(module, name, new_module)
            else:
                # recurse
                _replace(child)
    _replace(model)
    return model



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import torch


from .fn import *


class OMat24sAlexMPtrjLoss(torch.nn.Module):
    def __init__(
        self,
        energy_weight=1.0,
        forces_weight=1.0,
        stress_weight=1.0,
        energy_huber_delta=0.01,
        forces_huber_delta=0.01,
        stress_huber_delta=0.01,
        normalize: bool = False,
        **kwargs,
    ) -> None:
        "Modify code From MACE, see same logic as NequIP"
        super().__init__()
        if normalize:
            normalizer = energy_weight + forces_weight + stress_weight
            self.register_buffer(
                "energy_weight",
                torch.tensor(energy_weight / normalizer , dtype=torch.get_default_dtype()),
            )
            self.register_buffer(
                "forces_weight",
                torch.tensor(forces_weight / normalizer, dtype=torch.get_default_dtype()),
            )
            self.register_buffer(
                "stress_weight",
                torch.tensor(stress_weight / normalizer, dtype=torch.get_default_dtype()),
            )
        else:
            self.register_buffer(
                "energy_weight",
                torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
            )
            self.register_buffer(
                "forces_weight",
                torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
            )
            self.register_buffer(
                "stress_weight",
                torch.tensor(stress_weight, dtype=torch.get_default_dtype()),
            )
        self.energy_huber_delta = energy_huber_delta
        self.forces_huber_delta = forces_huber_delta
        self.stress_huber_delta = stress_huber_delta

    def forward(self, pred, label) -> Tensor:
        num_atoms = (label.ptr[1:] - label.ptr[:-1])
        batch = label["batch"]
        energy_weight = label.energy_weight
        forces_weight = label.forces_weight[batch].unsqueeze(-1)
        stress_weight = label.stress_weight.unsqueeze(-1).unsqueeze(-1)

        loss_energy = torch.nn.functional.huber_loss(
            energy_weight * label["energy"] / num_atoms,
            energy_weight * pred["energy"] / num_atoms,
            reduction="mean",
            delta=self.energy_huber_delta,
        )
        loss_forces = conditional_huber_forces(
            forces_weight * pred["forces"],
            forces_weight * label["forces"],
            huber_delta=self.forces_huber_delta,
        )
        loss_stress = torch.nn.functional.huber_loss(
            stress_weight * label["stress"],
            stress_weight * pred["stress"],
            reduction="mean",
            delta=self.stress_huber_delta,
        )
        return (
            self.energy_weight * loss_energy
            + self.forces_weight * loss_forces
            + self.stress_weight * loss_stress
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(\n"
            f"  energy_weight        = {self.energy_weight:.3f},\n"
            f"  forces_weight        = {self.forces_weight:.3f},\n"
            f"  stress_weight        = {self.stress_weight:.3f},\n"
            f"  energy_huber_delta   = {self.energy_huber_delta:.3f},\n"
            f"  forces_huber_delta   = {self.forces_huber_delta:.3f},\n"
            f"  stress_huber_delta   = {self.stress_huber_delta:.3f},\n"
            f")"
        )


class AQCat25OC20Loss(torch.nn.Module):
    def __init__(
        self,
        energy_weight=1.0,
        forces_weight=1.0,
        energy_huber_delta=0.01,
        forces_huber_delta=0.01,
        normalize: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()
        if normalize:
            normalizer = energy_weight + forces_weight
            self.register_buffer(
                "energy_weight",
                torch.tensor(energy_weight / normalizer , dtype=torch.get_default_dtype()),
            )
            self.register_buffer(
                "forces_weight",
                torch.tensor(forces_weight / normalizer, dtype=torch.get_default_dtype()),
            )
        else:
            self.register_buffer(
                "energy_weight",
                torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
            )
            self.register_buffer(
                "forces_weight",
                torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
            )
        self.energy_huber_delta = energy_huber_delta
        self.forces_huber_delta = forces_huber_delta

    def forward(self, pred, label) -> Tensor:
        num_atoms = (label.ptr[1:] - label.ptr[:-1])
        batch = label["batch"]
        energy_weight = label.energy_weight
        forces_weight = label.forces_weight[batch].unsqueeze(-1)

        loss_energy = torch.nn.functional.huber_loss(
            energy_weight * label["energy"] / num_atoms,
            energy_weight * pred["energy"] / num_atoms,
            reduction="mean",
            delta=self.energy_huber_delta,
        )
        loss_forces = conditional_huber_forces(
            forces_weight * pred["forces"],
            forces_weight * label["forces"],
            huber_delta=self.forces_huber_delta,
        )

        return (
            self.energy_weight * loss_energy
            + self.forces_weight * loss_forces
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(\n"
            f"  energy_weight        = {self.energy_weight:.3f},\n"
            f"  forces_weight        = {self.forces_weight:.3f},\n"
            f"  energy_huber_delta   = {self.energy_huber_delta:.3f},\n"
            f"  forces_huber_delta   = {self.forces_huber_delta:.3f},\n"
            f")"
        )



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import importlib
from typing import Any, Dict, Optional, List


import torch


from ..dataset.statistics import Statistics
from ..utils.utils import deep_convert


def select_wrapper(model_config: Dict) -> Any:
    wrapper_path = model_config.get("wrapper", {}).get("_target_", "tace.models.WrapModelV1")
    module_name, class_name = wrapper_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    wrap_cls = getattr(module, class_name)
    return wrap_cls


def select_model(
    cfg: Dict,
    statistics: Optional[Statistics],
    target_property: List[str],
    embedding_property: List[str],
    **kwargs,
) -> torch.nn.Module:
    
    if "model" in cfg:
        model_config = (cfg['model']['config'])
    else:
        model_config = cfg

    # === wrapper cls ===
    WRAPPER_CLS = select_wrapper(model_config)

    # === model cls ===
    model_path = model_config.get('_target_', 'tace.models.TACEV1')
    module_name, class_name = model_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    MODEL_CLS = getattr(module, class_name)
    model_config = deep_convert(model_config)
    # === instantiate ===
    try:
        MODEL = WRAPPER_CLS(
            MODEL_CLS(
                **model_config,
                target_property=target_property,
                embedding_property=embedding_property,
                statistics=statistics,
                model_config=model_config,
            )
        )
    except Exception as e:
        raise RuntimeError(
            f"Failed to instantiate the model using the provided configuration.\n"
            f"Model config: {model_config}"
        ) from e

    return MODEL



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from __future__ import annotations
from copy import deepcopy
from typing import Optional, List, Dict


import ase
import numpy as np
import torch
from torch_geometric.data import Data


from .element import TorchElement
from .neighbour_list import get_neighborhood
from .quantity import (
    KeySpecification,
    PROPERTY,
    UNIVERSAL_EMBEDDING_ALLOWED_PROPERTY,
)


def to_tensor(data):
    if data is None:
        raise ValueError(f"{data} must not be None")
    return torch.tensor(data, dtype=torch.get_default_dtype())


def build_adjacency_matrix_from_edge(edge_src, edge_dst, num_atoms):
    adjacency_matrix = torch.zeros((num_atoms, num_atoms), dtype=torch.int64)
    for source, target in zip(edge_src, edge_dst):
        adjacency_matrix[source, target] = 1
    return adjacency_matrix.reshape(1, num_atoms, num_atoms)


def from_atoms(
    element: TorchElement,
    atoms: ase.Atoms,
    cutoff: float,
    max_neighbors: Optional[int],
    keyspec: KeySpecification,
    target_property: List[str],
    embedding_property: List[str] = [],
    universal_embedding: Optional[List[Dict[str, Dict[int, str]]]] = None,
    training: bool = True,
    neighborlist_backend: str = "matscipy",
):
    # === The basic structure of chemical substances ===
    try:
        atomic_numbers = atoms.get_atomic_numbers()
    except Exception as e:
        raise RuntimeError(f"Failed to get atomic numbers from atoms: {e}")

    try:
        pbc = tuple(atoms.get_pbc())
    except Exception as e:
        raise RuntimeError(f"Failed to get pbc from atoms: {e}")

    try:
        lattice = np.array(atoms.get_cell())
    except Exception as e:
        raise RuntimeError(f"Failed to get cell from atoms: {e}")

    try:
        positions = atoms.get_positions()
    except Exception as e:
        raise RuntimeError(f"Failed to get positions from atoms: {e}")

    edge_index, edge_shifts, pbc, lattice = get_neighborhood(
        positions=positions,
        cutoff=cutoff,
        pbc=deepcopy(pbc),
        lattice=deepcopy(lattice),
        max_neighbors=max_neighbors,
        backend=neighborlist_backend,
    )

    
    atomic_numbers = torch.tensor(atomic_numbers, dtype=torch.int64)
    onehot = element.z2onehot(atomic_numbers).to(dtype=torch.get_default_dtype())

    # === Physical property to predict ===
    properties = {}
    property_weights = {}

    for name in list(keyspec.arrays_keys) + list(keyspec.info_keys):
        property_weights[name] = atoms.info.get(f"{name}_weight", 1.0)

    for name, atoms_key in keyspec.info_keys.items():
        properties[name] = atoms.info.get(atoms_key, None)
        if atoms_key not in atoms.info:
            property_weights[name] = 0.0

    for name, atoms_key in keyspec.arrays_keys.items():
        properties[name] = atoms.arrays.get(atoms_key, None)
        if atoms_key not in atoms.arrays:
            property_weights[name] = 0.0

    lattice = (
        torch.tensor(lattice.reshape((1, 3, 3)), dtype=torch.get_default_dtype())
        if lattice is not None
        else torch.tensor(3 * [0.0, 0.0, 0.0], dtype=torch.get_default_dtype()).view(
            1, 3, 3
        )
    )

    num_atoms = len(atomic_numbers)

    # === predict or embedding ===
    if training:
        need_property = set(target_property + embedding_property)
    else:
        need_property = set(embedding_property)

    joint_property = []
    for name in need_property:
        must_be_with = PROPERTY[name]['must_be_with']
        for _, v in must_be_with.items():
            joint_property += v
    need_property.update(list(set(joint_property)))

    pDict = {}
    wDict = {}
    for name in need_property:
        in_data = PROPERTY[name]["shape"]["in_data"]
        shape_fn = PROPERTY[name]["shape"].get("shape_fn", None)
        default_value_fn = PROPERTY[name]["default_value_fn"]
        type_ = PROPERTY[name]["type"]
        try:
            p = properties.get(name)
            if p is None:
                if type_ == 'float':
                    p = torch.tensor(
                        default_value_fn(num_atoms, type_),
                        dtype=torch.get_default_dtype(),
                    )
                elif type_ == 'int':
                    p = torch.tensor(
                        np.round(default_value_fn(num_atoms, type_)),
                        dtype=torch.int64,
                    )
                else:
                    raise
            else:
                if type_ == 'float':
                    p = torch.tensor(p, dtype=torch.get_default_dtype())
                elif type_ == 'int':
                    p = torch.tensor(np.round(p), dtype=torch.int64)
                else:
                    raise
                if shape_fn is not None:
                    p = shape_fn(
                        p,
                        num_atoms=num_atoms,
                    )
            pDict.update({name: p.view(*in_data)})
        except Exception as e:
            raise RuntimeError(f"Failed to read property ``{name}``") from e

        try:
            w = (
                torch.tensor(
                    property_weights.get(name), dtype=torch.get_default_dtype()
                ).view(
                    1
                )  # (1,)
                if property_weights.get(name) is not None
                else torch.tensor(1.0, dtype=torch.get_default_dtype())  # ()
            )
            wDict.update({name: w})
        except Exception as e:
            raise RuntimeError(f"Failed to read property ``{name}``") from e

    if universal_embedding is not None:
        for p in embedding_property:
            assert (
                p in UNIVERSAL_EMBEDDING_ALLOWED_PROPERTY
            ), f"universal_embedding key allowed property are {UNIVERSAL_EMBEDDING_ALLOWED_PROPERTY}, "
            f"if you need more, please contact the author"

    data_dict = {
        "entropy": to_tensor(atoms.info.get('entropy', 1.0)),
        "atomic_numbers": atomic_numbers,
        "lattice": lattice,
        "positions": to_tensor(positions),
        "node_attrs": onehot,
        "edge_index": torch.tensor(edge_index, dtype=torch.int64),
        "edge_shifts": to_tensor(edge_shifts),
        "level": torch.tensor(atoms.info.get(keyspec.info_keys['level'], 0), dtype=torch.int64),
    }

    for name in need_property:
        data_dict.update(
            {
                name: pDict[name],
                f"{name}_weight": wDict[name],
            }
        )
    return Data(**data_dict)



from .v1.tace import TACEV1
from .v1.wrapper import WrapModelV1

__all__ = [
    "TACEV1",
    "WrapModelV1",
]

try:
    from .v2.tace import TACEV2
    from .v2.wrapper import WrapModelV2

    __all__.extend([
        "TACEV2",
        "WrapModelV2",
    ])
except ImportError:
    pass



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################


import argparse

import torch


from ..lightning import load_tace
from ..utils._global import DTYPE


ALLOWED_BACKEND = ["state_dict", "whole_model", "lammps"]


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-m", "--model",
        type=str,
        required=True,
        help="Model path",
    )
    parser.add_argument(
        "-l", "--level",
        type=int,
        default=None,
        help="Which fidelity to use",
    )
    parser.add_argument(
        "--dtype",
        type=str,
        help="Model dtype",
        choices=['float32', 'float64'],
        default=None,
    )
    parser.add_argument(
        "--device", 
        type=str, 
        default="cuda",
        choices=["cpu", "cuda"], 
        help="Device for inference"
    )
    parser.add_argument(
        "--backend", 
        type=str, 
        default="state_dict",
        choices=ALLOWED_BACKEND, 
        help="Specify the backend to export"
    )
    return parser.parse_args()


def main():
    args = parse_args()
    model = load_tace(args.model, args.device, strict=True, use_ema=True)
    model_dtype = model.readout_fn.cutoff.dtype
    args_dtype = DTYPE[args.dtype] or model_dtype
    if args_dtype != model_dtype:
        print(f"[Warning] Model dtype does not match args.dtype. Forcing dtype from {model_dtype} to {args_dtype}")
    torch.set_default_dtype(args_dtype)
    if args.level is not None:
        model.level = args.level
    model.to(dtype=args_dtype, device=args.device)

    if args.backend == "state_dict":
        torch.save(
            {
                "state_dict": model.state_dict(),
                "cfg": model.readout_fn.model_config,
                "target_property": model.readout_fn.target_property,
                "embedding_property": model.readout_fn.embedding_property,
                "statistics": model.readout_fn.statistics,
            }, 
            args.model + "-state.pt"
        )
    elif args.backend == "whole_model":
        torch.save(model, args.model + "-whole.pt") 
    elif args.backend == "lammps":
        from ..interface.lammps import TACELammpsCalc
        model.lmp = True
        lammps_model = TACELammpsCalc(model)
        torch.save(lammps_model, args.model + "-lammps_mliap.pt")
    else:
        raise ValueError(f"Unsupported backend '{args.backend}'. One of {ALLOWED_BACKEND} is available.")

if __name__ == "__main__":
    main()


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
'''
This interface is general. As long as you have successfully compiled LAMMPS 
with any one of the models TACE, MACE, NequIP, or Allegro ...
then all MLIPs that have an interface with LAMMPS-ML-IAP can be used.
'''


import logging
from typing import Dict, Tuple

import torch
from torch import Tensor
from ase.data import chemical_symbols
from lammps.mliap.mliap_unified_abc import MLIAPUnified


from ...lightning import load_tace

# TODO check device 
class EdgeForcesWrapper(torch.nn.Module):
    def __init__(self, model: torch.nn.Module, **kwargs):
        super().__init__()
        model.lmp = True
        model.flags.compute_forces = False
        model.flags.compute_stress = False
        model.flags.compute_virials = False
        model.flags.compute_edge_forces = True
        
        self.model = model
        self.register_buffer("cutoff", model.readout_fn.cutoff)
        self.register_buffer("atomic_numbers", model.readout_fn.atomic_numbers)
        self.register_buffer("num_layers", model.readout_fn.num_layers)
        for p in self.model.parameters():
            p.requires_grad = False

    def forward(
        self, data: Dict[str, Tensor]
    ) -> Tuple[Tensor, Tensor, Tensor]:

        outs = self.model(data)
        node_energy = outs["node_energy"]
        pair_forces = outs["edge_forces"]
        total_energy = outs["energy"][0]

        if pair_forces is None:
            pair_forces = torch.zeros_like(data["edge_vector"])

        return total_energy, node_energy, pair_forces


class TACELammpsCalc(MLIAPUnified):
    '''Not test for cpu running, only cuda devices are tested by author'''
    def __init__(self, model, **kwargs):
        super().__init__()
        model = load_tace(model)
        self.model = EdgeForcesWrapper(model, **kwargs)
        self.element_types = [chemical_symbols[s] for s in model.readout_fn.atomic_numbers]
        self.num_species = len(self.element_types)
        self.rcutfac = 0.5 * float(model.readout_fn.cutoff)
        self.nparams = 1
        self.ndescriptors = 1

        self.dtype = model.readout_fn.cutoff.dtype
        self.device = "cpu"
        self.initialized = False

    def _initialize_device(self, data):
        device = torch.as_tensor(data.elems).device
        self.device = device
        # self.device = (
        #     "cuda" if "kokkos" in data.__class__.__module__.lower() else "cpu"
        # )
        self.model = self.model.to(device)
        logging.info(f"TACE model initialized on device: {device}")
        self.initialized = True

    def compute_forces(self, data):
        nlocal = data.nlocal
        ntotal = data.ntotal
        npairs = data.npairs
        nghosts = ntotal - nlocal
        species = torch.as_tensor(data.elems, dtype=torch.int64)

        if not self.initialized:
            self._initialize_device(data)

        if nlocal == 0 or npairs <= 1:
            return
            
        batch = self._prepare_batch(data, nlocal, nghosts, species)

        _, node_energy, pair_forces = self.model(batch)
        if self.device.type != "cpu":
            torch.cuda.synchronize()

        self._update_lammps_data(data, node_energy, pair_forces, nlocal)

    def _prepare_batch(self, data, nlocal, nghosts, species):
        edge_vector = torch.as_tensor(data.rij).to(self.dtype).to(self.device)
        edge_vector.requires_grad_(True)
        return {
            "edge_vector": edge_vector,
            "node_attrs": torch.nn.functional.one_hot(
                species.to(self.device), num_classes=self.num_species
            ).to(self.dtype),
            "edge_index": torch.stack(
                [
                    torch.as_tensor(data.pair_j, dtype=torch.int64).to(self.device),
                    torch.as_tensor(data.pair_i, dtype=torch.int64).to(self.device),
                ],
                dim=0,
            ),
            "batch": torch.zeros(nlocal, dtype=torch.int64, device=self.device),
            "ptr": torch.tensor([0, nlocal], dtype=torch.int64, device=self.device),
            "lmp_data": data,
            "natoms": (nlocal, nghosts),
        }


    def _update_lammps_data(self, data, node_energy, pair_forces, nlocal):
        if self.dtype == torch.float32:
            pair_forces = pair_forces.double()
            node_energy = node_energy.double()
        eatoms = torch.as_tensor(data.eatoms)
        eatoms.copy_(node_energy[:nlocal])
        data.energy = torch.sum(node_energy[:nlocal]).detach()
        # print(torch.sum(node_energy[:natoms]))
        # import sys
        # sys.exit()
        data.update_pair_forces_gpu(pair_forces)

    def compute_descriptors(self, data):
        pass

    def compute_gradients(self, data):
        pass


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import List, Dict, Optional


import torch
from torch import Tensor
from e3nn.util.jit import compile_mode

from .tace import TACEV1
from .wrapper_utils import (
    compute_symmetric_displacement, 
    compute_atomic_virials_stresses,
    compute_hessians_vmap,
    sample_force_jacobian,
)
from .utils import Graph
from ...dataset.quantity import PROPERTY, ComputeFlag
from ...utils.torch_scatter import scatter_sum


@compile_mode("script")
class WrapModelV1(torch.nn.Module):

    level: int
    spin_on: int

    def __init__(self, readout_fn: TACEV1):
        super().__init__()
        self.readout_fn = readout_fn
        target_property = readout_fn.target_property
        self._set_target_property(target_property)
        self._set_lammps_mliap()
        self.set_embedding_property()
        self.set_universal_embedding()
        self._set_spin_on()
        self._set_computing_level()
   
    def _set_spin_on(self, enable: bool = False):
        self.spin_on = 1 if enable else 0

    def reset_spin_on(self, enable: bool = True):
        assert isinstance(enable, bool)
        self._set_spin_on(enable)

    def get_spin_on(self) -> int:
        return self.spin_on

    def set_embedding_property(self):
        self.embedding_property = getattr(
            self.readout_fn, "embedding_property", []
        )

    def set_universal_embedding(self):
        self.universal_embedding = getattr(
            self.readout_fn, "universal_embedding", {}
        )

    def _set_computing_level(self, level: int = 0):
        self.level = level
        
    def reset_computing_level(self, level: int = 0):
        assert isinstance(level, int) and level >= 0
        self.level = level

    def get_computing_level(self) -> int:
        return self.level

    def _set_target_property(self, target_property: List[str]):

        if 'direct_hessians' in target_property:
            raise
        if 'final_collinear_magmoms' in target_property:
            raise
        if 'final_noncollinear_magmoms' in target_property:
            raise

        self.target_property = target_property

        self.flags = ComputeFlag()
        for k in self.target_property:
            setattr(self.flags, f"compute_{k}", k in self.target_property)

        self.compute_first_derivative = False
        for p in self.target_property:
            if PROPERTY[p]['first_derivative']:
                self.compute_first_derivative = True

        self.compute_second_derivative = False
        for p in self.target_property:
            if PROPERTY[p]['second_derivative']:
                self.compute_second_derivative = True

        self.retain_graph = self.compute_second_derivative
        self.create_graph = self.compute_second_derivative\
        
    def reset_target_property(self, target_property: List[str]):
        assert isinstance(target_property, List)
        self._set_target_property(target_property)
        self.readout_fn._reset_target_property(target_property)

    def get_target_property(self) -> List[str]:
        return self.target_property

    def _set_lammps_mliap(self, enable: bool = False):
        self.lmp = enable

    def reset_lammps_mliap(self, enable: bool = True):
        assert isinstance(enable, bool)
        self._set_lammps_mliap(enable)

    def forward(self, data: Dict[str, torch.Tensor]) -> Dict[str, Optional[torch.Tensor | List[torch.Tensor]]]:
        # === pre processing ===
        graph = self.prepare_graph(data)

        # === predict ===
        RESULTS = self.readout_fn(data, graph)
        FIRST = self.first_derivative_fn(data, graph, RESULTS)
        SECOND = self.second_derivative_fn(data, graph, RESULTS, FIRST)

        return {**RESULTS, **FIRST, **SECOND}


    def first_derivative_fn(
        self, data: Dict[str, Tensor], graph: Graph, results: Dict[str, Tensor]
    ) -> Dict[str, Optional[Tensor]]:

        E = results["energy"]
        F = None
        V = None
        S = None
        P = None
        M = None
        C_MAG_F = None
        NC_MAG_F = None
        EDGE_F = None
        A_V = None
        A_S = None

        inputs = []
        if self.flags.compute_forces:
            inputs.append(graph.positions)
        if self.flags.compute_stress or self.flags.compute_virials:
            inputs.append(graph.displacement)
        if self.flags.compute_polarization or self.flags.compute_conservative_dipole:
            inputs.append(data["electric_field"])
        if self.flags.compute_magnetization:
            inputs.append(data["magnetic_field"])
        if self.flags.compute_collinear_magnetic_forces:
            inputs.append(data["initial_collinear_magmoms"])
        if self.flags.compute_noncollinear_magnetic_forces:
            inputs.append(data["initial_noncollinear_magmoms"])
        if self.flags.compute_edge_forces or \
            self.flags.compute_atomic_virials or \
            self.flags.compute_atomic_stresses:
            inputs.append(graph.edge_vector)
            
        if self.compute_first_derivative:
            grad_outputs = torch.ones_like(E)
            grads = torch.autograd.grad(
                outputs=E,
                inputs=inputs,
                grad_outputs=grad_outputs,
                retain_graph=self.training or self.retain_graph,
                create_graph=self.training or self.create_graph,
                allow_unused=True,
            )
        idx = 0
        if self.flags.compute_forces:
            F = -grads[idx]
            idx += 1
        if self.flags.compute_stress or self.flags.compute_virials:
            V = -grads[idx]
            VOLUME = torch.linalg.det(data["lattice"]).abs().unsqueeze(-1)
            S = -V / VOLUME.view(-1, 1, 1)
            S = torch.where(torch.abs(S) < 1e10, S, torch.zeros_like(S))
            idx += 1
        if self.flags.compute_polarization or self.flags.compute_conservative_dipole:
            P = -grads[idx]
            idx += 1
        if self.flags.compute_magnetization:
            M = -grads[idx]
            idx += 1
        if self.flags.compute_collinear_magnetic_forces:
            C_MAG_F = -grads[idx]
            idx += 1
        if self.flags.compute_noncollinear_magnetic_forces:
            NC_MAG_F = -grads[idx]
            idx += 1
        if self.flags.compute_edge_forces or \
            self.flags.compute_atomic_virials or \
            self.flags.compute_atomic_stresses:
            EDGE_F = grads[idx] # consistency with LAMMPS
            idx += 1
            if not self.lmp:
                A_V, A_S = compute_atomic_virials_stresses(
                    graph.edge_vector,
                    EDGE_F,
                    data['edge_index'],
                    graph.lattice,
                    data['batch'],
                    data['node_attrs'].size(0),
                    True,
                    True,
                )
                
        return {
            "forces": F,
            "virials": V,
            "stress": S,
            "polarization": P,
            "magnetization": M,
            "collinear_magnetic_forces": C_MAG_F,
            "noncollinear_magnetic_forces": NC_MAG_F,
            "edge_forces": EDGE_F,
            "atomic_virials": A_V,
            "atomic_stresses": A_S,
        }

    def second_derivative_fn(
        self,
        data: Dict[str, torch.Tensor],
        graph: Graph,
        extra: Dict[str, torch.Tensor],
        first_derivative: Dict[str, torch.Tensor],
    ) -> Dict[str, torch.Tensor]:

        forces = first_derivative["forces"]
        polarization = first_derivative["polarization"]
        magnetization = first_derivative["magnetization"]

        inputs_list: List[Tensor] = []
        if self.flags.compute_born_effective_charges:
            inputs_list.append(graph.positions)
        if self.flags.compute_conservative_polarizability:
            inputs_list.append(data["electric_field"])

        BEC = None
        ALPHA = None
        CHI_M = None
        HESSIANS = None
        jacs_per_graph = None
        samples_per_graph = None

        BECList = []
        ALPHAList = []
        CHI_MList = []
        if self.flags.compute_conservative_polarizability or self.flags.compute_born_effective_charges:
            for i in range(3):  # μ = 0,1,2

                polarization_i = polarization.sum(dim=0)[i]  # sum over batch dimension
                grads = torch.autograd.grad(
                    outputs=polarization_i,
                    inputs=inputs_list,
                    retain_graph=(
                        i != 2 
                        or self.training 
                        or self.flags.compute_hessians 
                        or self.flags.compute_magnetization
                    ),
                    create_graph=self.training,
                )
                idx = 0
                if self.flags.compute_born_effective_charges:
                    BEC = grads[idx]
                    idx += 1
                if self.flags.compute_conservative_polarizability:
                    ALPHA = grads[idx]

                if BEC is None:
                    BEC = torch.zeros(
                        graph.positions.shape[0],
                        3,
                        3,
                        device=graph.positions.device,
                        dtype=graph.positions.dtype,
                    )
                if ALPHA is None:
                    ALPHA = torch.zeros_like(data["electric_field"])

                BECList.append(BEC)  # [atoms, 3]
                ALPHAList.append(ALPHA)  # [B,3]

            BEC = torch.stack(BECList, dim=0)  # [3, atoms, 3]
            BEC = BEC.transpose(1, 0)  #        # [atoms, 3 (pol), 3 (pos)]
            ALPHA = torch.stack(ALPHAList, dim=1)  # [B,3,3]

        if self.flags.compute_magnetization:
            MF = data["magnetic_field"]
            for i in range(3):
                mag_i = magnetization.sum(dim=0)[i]
                CHI_M = torch.autograd.grad(
                    outputs=mag_i,
                    inputs=[MF],
                    retain_graph=(
                        i != 2 
                        or self.training 
                        or self.flags.compute_hessians 
                    ),
                    create_graph=self.training,
                    allow_unused=True,
                )[0]
                if CHI_M is None:
                    CHI_M = torch.zeros_like(MF)
                CHI_MList.append(CHI_M)  # [B,3]
            CHI_M = torch.stack(CHI_MList, dim=1)  # [B,3,3]


        if self.flags.compute_hessians:
            if self.training:
                jacs_per_graph, samples_per_graph = sample_force_jacobian(
                    forces, 
                    graph.positions, 
                    data["ptr"], 
                    num_samples=2,
                    create_graph=self.training,
                )
            else:
                HESSIANS = compute_hessians_vmap(forces, graph.positions)

        return {
            "hessians": HESSIANS,
            "jacs_per_graph": jacs_per_graph,
            "samples_per_graph": samples_per_graph,
            "conservative_polarizability": ALPHA,
            "born_effective_charges": BEC,
            "magnetic_susceptibility": CHI_M,
        }

    def prepare_graph(self, data: Dict[str, torch.Tensor]) -> Graph:

        node_level = (
            data['level'][data['batch']]
            if "level" in data
            else torch.full_like(data['batch'], self.level, dtype=torch.int64)
        )  # used for multi-fidelity and multi-head

        if "spin_on" not in data: # used for uie
            data['spin_on'] = torch.tensor([self.spin_on], device=data["node_attrs"].device, dtype=torch.int64)

        if self.lmp:
            for p in self.target_property:
                for requires_grad_p in PROPERTY[p]['requires_grad_with']:
                    if p != 'forces':
                        data[requires_grad_p].requires_grad_(True)
            dtype = data["node_attrs"].dtype
            device =  data["node_attrs"].device 
            nlocal, nghosts = data["natoms"][0], data["natoms"][1]
            num_graphs = 2
            displacement = None
            positions = torch.zeros(
                (int(nlocal), 3),
                dtype=dtype,
                device=device,
            )
            lattice = torch.zeros(
                (num_graphs, 3, 3),
                dtype=dtype,
                device=device,
            )
            edge_vector = data["edge_vector"].requires_grad_(True)
            edge_length = (edge_vector**2).sum(dim=1, keepdim=True).sqrt() + 1e-9
            lmp_data = data["lmp_data"]
            lmp_natoms = (nlocal, nghosts)
            num_atoms_arange = torch.arange(nlocal, device=positions.device, dtype=torch.int64)
        else:
            requires_grad_p_list = []
            for p in self.target_property:
                for requires_grad_p in PROPERTY[p]['requires_grad_with']:
                    requires_grad_p_list.append(requires_grad_p)
                    if requires_grad_p != "edge_vector":
                        data[requires_grad_p].requires_grad_(True)
            dtype = data["node_attrs"].dtype
            device =  data["node_attrs"].device 
            positions = data["positions"]
            num_graphs = data["ptr"].numel() - 1
            displacement = torch.zeros(
                (num_graphs, 3, 3), dtype=dtype, device=device
            )
            if self.flags.compute_virials or self.flags.compute_stress:
                displacement = compute_symmetric_displacement(data, num_graphs)
            source = data["edge_index"][0]
            target = data["edge_index"][1]
            edge_batch = data["batch"][source]
            edge_vector = (
                data["positions"][target]
                - data["positions"][source]
                + torch.einsum(
                    "ni,nij->nj", data["edge_shifts"], data["lattice"][edge_batch]
                )
            )
            if set(self.target_property) & {"edge_vector", "atomic_stresses", "atomic_virials"}:
                    edge_vector.requires_grad_(True)
            edge_length = (edge_vector**2).sum(dim=1, keepdim=True).sqrt() + 1e-9
            lattice = data['lattice']
            lmp_data = None
            lmp_natoms = (positions.size(0), 0)
            num_atoms_arange = torch.arange(positions.shape[0], device=positions.device, dtype=torch.int64)

        return Graph(
            lmp=self.lmp,
            lmp_data=lmp_data,
            lmp_natoms=lmp_natoms, 
            num_graphs=num_graphs,
            displacement=displacement,
            positions=positions,
            edge_vector=edge_vector,
            edge_length=edge_length,
            lattice=lattice,
            node_level=node_level,
            num_atoms_arange=num_atoms_arange,
        )
    





################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import argparse
import yaml


from ..lightning import load_tace


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--model", type=str, required=True, help="Path to model checkpoint (.ckpt or .pt or .pth)")
    return parser.parse_args()


def main():
    args = parse_args()
    model = load_tace(args.model, 'cpu', strict=True, use_ema=True)
    finetune_cfg = {}
  
    # === LoRA ===
    finetune_cfg['lora'] = {}
    def set_lora_config(name: str, r: int, alpha: float):
        finetune_cfg['lora'][name] = {
            'r': int(r),
            'alpha': float(alpha),
        }
    set_lora_config('element_embedding', 4, 8)
    set_lora_config('radial_mlp', 4, 8)
    set_lora_config('interaction', 4, 8)
    set_lora_config('product', 4, 8)
    set_lora_config('readout', 4, 8)

    # === Freeze ===
    finetune_cfg['freeze'] = {}
    for name, _ in model.named_parameters():
        finetune_cfg['freeze'][name] = True

    with open('finetune_config.yaml', "w") as f:
        yaml.dump(finetune_cfg, f, default_flow_style=False, sort_keys=False)

    print(
        "The finetune_config.yaml file has been created. tace-train will automatically read this file. "
        "You can modify the parameters in it to control the degree of freedom for fine-tuning. "
        "By default, all parameters are frozen and LoRA is used for fine-tuning. "
        "After training is completed, it is recommended to use the tace-convert command "
        "to merge the LoRA parameters with the base model's parameters."
    )

if __name__ == "__main__":
    main()



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from math import sqrt
from typing import Optional, List


import torch
from torch import nn


from .act import ACT


class MLP(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        hidden_dim: List[int] = [],
        act: Optional[str] = "silu",
        bias: bool = False,
        forward_weight_init: bool = True,
        enable_layer_norm: bool = False,
    ):
        '''The parameter initialization method uses the earlier version of Allegro'''
        super().__init__()
        self.bias = bias
        self.dims = [in_dim] + hidden_dim + [out_dim]
        self.num_layers = len(self.dims) - 1
        assert self.num_layers >= 1
        act = ACT[act]()
        self.is_nonlinear = False
        self.enable_layer_norm = enable_layer_norm

        # === build the MLP + weight init ===
        mlp = []
        for layer, (h_in, h_out) in enumerate(zip(self.dims, self.dims[1:])):

            # === weight initialization ===
            if forward_weight_init:
                norm_dim = h_in
                gain = 1.0 if act is None or (layer == 0) else sqrt(2)
            else:
                norm_dim = h_out
                gain = (
                    1.0 if act is None or (layer == self.num_layers - 1) else sqrt(2)
                )

            # === instantiate Linear ===
            linear_layer = LinearLayer(
                in_dim=h_in,
                out_dim=h_out,
                alpha=gain / sqrt(norm_dim),
                bias=bias,
            )
            mlp.append(linear_layer)

            # === optional LayerNorm ===
            if enable_layer_norm:
                if layer < len(self.dims) -2:
                    mlp.append(nn.LayerNorm(h_out))
            del gain, norm_dim

            # === act ===
            if (layer != self.num_layers - 1) and (act is not None):
                mlp.append(act)
                self.is_nonlinear = True

        self.mlp = torch.nn.Sequential(*mlp)

    def forward(self, x):
        return self.mlp(x)


class LinearLayer(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        alpha: float = 1.0,
        bias: bool = False,
    ) -> None:
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.alpha = alpha
        self.weight = torch.nn.Parameter(torch.empty((in_dim, out_dim)))
        torch.nn.init.uniform_(self.weight, -sqrt(3), sqrt(3))
        self._bias = bias
        if bias:
            self.bias = torch.nn.Parameter(torch.zeros(out_dim))
        else:
            self.register_parameter("bias", None)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        weight = self.weight * self.alpha
        if self.bias is None:
            return torch.mm(input, weight)
        else:
            return torch.addmm(self.bias, input, weight)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(in_dim={self.in_dim}, out_dim={self.out_dim} bias={ self._bias})"


class LoRALinearLayer(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        alpha: float = 1.0,
        bias: bool = False,
        lora_r: int = 8,
        lora_alpha: float = 8.0,
        **kwargs,
    ) -> None:
        super().__init__()

        self.use_lora = True
        self.lora_alpha = lora_alpha
        self.lora_r = lora_r
        self.scaling = lora_alpha / lora_r

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.alpha = alpha
        self.weight = torch.nn.Parameter(torch.empty((in_dim, out_dim)))
        torch.nn.init.uniform_(self.weight, -sqrt(3), sqrt(3))
        self.weight.requires_grad_(False)
        if bias:
            self.bias = torch.nn.Parameter(torch.zeros(out_dim))
            self.bias.requires_grad_(False)
        else:
            self.register_parameter("bias", None)

        self.lora_A = nn.Parameter(torch.zeros(in_dim, lora_r))
        self.lora_B = nn.Parameter(torch.zeros(lora_r, out_dim))
        nn.init.kaiming_uniform_(self.lora_A, a=sqrt(5))
        nn.init.zeros_(self.lora_B)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.use_lora:
            delta_w = self.lora_A @ (self.lora_B * self.scaling)
            W = self.weight + delta_w
        else:
            W = self.weight

        W = W * self.alpha

        if self.bias is None:
            return torch.mm(input, W)
        else:
            return torch.addmm(self.bias, input, W)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"  in_dim      = {self.in_dim},\n"
            f"  out_dim     = {self.out_dim},\n"
            f"  bias        = {self.bias is not None},\n"
            f"  alpha       = {self.alpha:.2f},\n"
            f"  lora_r      = {self.lora_r},\n"
            f"  lora_alpha  = {self.lora_alpha},\n"
            f")"
        )
